"""
Modules for reading in NOAA GML data

"""
import re
from datetime import datetime
from pathlib import Path

import numpy as np
import xarray as xr

import act


def read_gml(filename, datatype=None, remove_time_vars=True, convert_missing=True, **kwargs):
    """
    Function to call or guess what reading NOAA GML daga routine to use. It tries to
    guess the correct reading function to call based on filename. It mostly
    works, but you may want to specify for best results.

    Parameters
    ----------
    filename : str or pathlib.Path
        Data file full path name. In theory it should work with a list of
        filenames but it is not working as well with that as expected.
    datatype : str
        Data file type that bypasses the guessing from filename format
        and goes directly to the reading routine. Options include
        [MET, RADIATION, OZONE, CO2, HALO]
    remove_time_vars : bool
        Some variables are convereted into coordinate variables in Xarray
        DataSet and not needed after conversion. This will remove those
        variables.
    convert_missing : bool
        Convert missing value indicator in CSV to NaN in Xarray DataSet.
    **kwargs : keywords
        Keywords to pass through to instrument specific reading routine.

    Returns
    -------
    dataset : Xarray.dataset
        Standard ARM Xarray dataset with the data cleaned up to have units,
        long_name, correct type and some other stuff.

    """

    if datatype is not None:
        if datatype.upper() == 'MET':
            return read_gml_met(filename, convert_missing=convert_missing, **kwargs)
        elif datatype.upper() == 'RADIATION':
            return read_gml_radiation(
                filename,
                remove_time_vars=remove_time_vars,
                convert_missing=convert_missing,
                **kwargs,
            )
        elif datatype.upper() == 'OZONE':
            return read_gml_ozone(filename, **kwargs)
        elif datatype.upper() == 'CO2':
            return read_gml_co2(filename, convert_missing=convert_missing, **kwargs)
        elif datatype.upper() == 'HALO':
            return read_gml_halo(filename, **kwargs)
        else:
            raise ValueError('datatype is unknown')

    else:
        test_filename = filename
        if isinstance(test_filename, (list, tuple)):
            test_filename = filename[0]

        test_filename = str(Path(test_filename).name)

        if test_filename.startswith('met_') and test_filename.endswith('.txt'):
            return read_gml_met(filename, convert_missing=convert_missing, **kwargs)

        if test_filename.startswith('co2_') and test_filename.endswith('.txt'):
            return read_gml_co2(filename, convert_missing=convert_missing, **kwargs)

        result = re.match(r'([a-z]{3})([\d]{5}).dat', test_filename)
        if result is not None:
            return read_gml_radiation(
                filename,
                remove_time_vars=remove_time_vars,
                convert_missing=convert_missing,
                **kwargs,
            )

        ozone_pattern = [
            r'[a-z]{3}_[\d]{4}_[\d]{2}_hour.dat',
            r'[a-z]{3}_[\d]{2}_[\d]{4}_hour.dat',
            r'[a-z]{3}_[\d]{4}_all_minute.dat',
            r'[a-z]{3}_[\d]{2}_[\d]{4}_5minute.dat',
            r'[a-z]{3}_[\d]{2}_[\d]{4}_min.dat',
            r'[a-z]{3}_o3_6m_hour_[\d]{2}_[\d]{4}.dat',
            r'[a-z]{3}_ozone_houry__[\d]{4}',
        ]
        for pattern in ozone_pattern:
            result = re.match(pattern, test_filename)
            if result is not None:
                return read_gml_ozone(filename, **kwargs)

        ozone_pattern = [
            r'[a-z]{3}_CCl4_Day.dat',
            r'[a-z]{3}_CCl4_All.dat',
            r'[a-z]{3}_CCl4_MM.dat',
            r'[a-z]{3}_MC_MM.dat',
        ]
        for pattern in ozone_pattern:
            result = re.match(pattern, test_filename)
            if result is not None:
                return read_gml_halo(filename, **kwargs)


def read_gml_halo(filename, **kwargs):
    """
    Function to read Halocarbon data from NOAA GML.

    Parameters
    ----------
    filename : str or pathlib.Path
        Data file full path name.

    Returns
    -------
    dataset : Xarray.dataset
        Standard ARM Xarray dataset with the data cleaned up to have units,
        long_name, correct type and some other stuff.
    **kwargs : keywords
        Keywords to pass through to ACT read_csv() routine.

    """

    ds = None
    if filename is None:
        return ds

    variables = {
        'CCl4catsBRWm': {
            'long_name': 'Carbon Tetrachloride (CCl4) daily median',
            'units': 'ppt',
            '_FillValue': np.nan,
            '__type': np.float32,
            '__rename': 'CCl4',
        },
        'CCl4catsBRWmsd': {
            'long_name': 'Carbon Tetrachloride (CCl4) standard deviation',
            'units': 'ppt',
            '_FillValue': np.nan,
            '__type': np.float32,
            '__rename': 'CCl4_std_dev',
        },
        'CCl4catsBRWsd': {
            'long_name': 'Carbon Tetrachloride (CCl4) standard deviation',
            'units': 'ppt',
            '_FillValue': np.nan,
            '__type': np.float32,
            '__rename': 'CCl4_std_dev',
        },
        'CCl4catsBRWn': {
            'long_name': 'Number of samples',
            'units': 'count',
            '__type': np.int16,
            '__rename': 'number_of_samples',
        },
        'CCl4catsBRWunc': {
            'long_name': 'Carbon Tetrachloride (CCl4) uncertainty',
            'units': 'ppt',
            '_FillValue': np.nan,
            '__type': np.float32,
            '__rename': 'CCl4_uncertainty',
        },
        'MCcatsBRWm': {
            'long_name': 'Methyl Chloroform (CH3CCl3)',
            'units': 'ppt',
            '_FillValue': np.nan,
            '__type': np.float32,
            '__rename': 'methyl_chloroform',
        },
        'MCcatsBRWunc': {
            'long_name': 'Methyl Chloroform (CH3CCl3) uncertainty',
            'units': 'ppt',
            '_FillValue': np.nan,
            '__type': np.float32,
            '__rename': 'methyl_chloroform_uncertainty',
        },
        'MCcatsBRWsd': {
            'long_name': 'Methyl Chloroform (CH3CCl3) standard deviation',
            'units': 'ppt',
            '_FillValue': np.nan,
            '__type': np.float32,
            '__rename': 'methyl_chloroform_std_dev',
        },
        'MCcatsBRWmsd': {
            'long_name': 'Methyl Chloroform (CH3CCl3) standard deviation',
            'units': 'ppt',
            '_FillValue': np.nan,
            '__type': np.float32,
            '__rename': 'methyl_chloroform_std_dev',
        },
        'MCcatsBRWn': {
            'long_name': 'Number of samples',
            'units': 'count',
            '__type': np.int16,
            '__rename': 'number_of_samples',
        },
        'MCritsBRWm': {
            'long_name': 'Methyl Chloroform (CH3CCl3)',
            'units': 'ppt',
            '_FillValue': np.nan,
            '__type': np.float32,
            '__rename': 'methyl_chloroform',
        },
        'MCritsBRWsd': {
            'long_name': 'Methyl Chloroform (CH3CCl3) standard deviation',
            'units': 'ppt',
            '_FillValue': np.nan,
            '__type': np.float32,
            '__rename': 'methyl_chloroform_std_dev',
        },
        'MCritsBRWn': {
            'long_name': 'Number of samples',
            'units': 'count',
            '__type': np.int16,
            '__rename': 'number_of_samples',
        },
    }

    test_filename = filename
    if isinstance(test_filename, (list, tuple)):
        test_filename = test_filename[0]

    with open(test_filename) as fc:
        header = 0
        while True:
            line = fc.readline().strip()
            if not line.startswith('#'):
                break
            header += 1

    ds = act.io.csvfiles.read_csv(
        filename, sep=r'\s+', header=header, na_values=['Nan', 'NaN', 'nan', 'NAN'], **kwargs
    )
    var_names = list(ds.data_vars)
    year_name, month_name, day_name, hour_name, min_name = None, None, None, None, None
    for var_name in var_names:
        if var_name.endswith('yr'):
            year_name = var_name
        elif var_name.endswith('mon'):
            month_name = var_name
        elif var_name.endswith('day'):
            day_name = var_name
        elif var_name.endswith('hour'):
            hour_name = var_name
        elif var_name.endswith('min'):
            min_name = var_name

    timestamp = np.full(ds[var_names[0]].size, np.nan, dtype='datetime64[s]')
    for ii in range(0, len(timestamp)):
        if min_name is not None:
            ts = datetime(
                ds[year_name].values[ii],
                ds[month_name].values[ii],
                ds[day_name].values[ii],
                ds[hour_name].values[ii],
                ds[min_name].values[ii],
            )
        elif hour_name is not None:
            ts = datetime(
                ds[year_name].values[ii],
                ds[month_name].values[ii],
                ds[day_name].values[ii],
                ds[hour_name].values[ii],
            )
        elif day_name is not None:
            ts = datetime(
                ds[year_name].values[ii],
                ds[month_name].values[ii],
                ds[day_name].values[ii],
            )
        else:
            ts = datetime(ds[year_name].values[ii], ds[month_name].values[ii], 1)

        timestamp[ii] = np.datetime64(ts)

    for var_name in [year_name, month_name, day_name, hour_name, min_name]:
        try:
            del ds[var_name]
        except KeyError:
            pass

    ds = ds.rename({'index': 'time'})
    ds = ds.assign_coords(time=timestamp)
    ds['time'].attrs['long_name'] = 'Time'

    for var_name, value in variables.items():
        if var_name not in var_names:
            continue

        for att_name, att_value in value.items():
            if att_name == '__type':
                values = ds[var_name].values
                values = values.astype(att_value)
                ds[var_name].values = values
            elif att_name == '__rename':
                ds = ds.rename({var_name: att_value})
            else:
                ds[var_name].attrs[att_name] = att_value

    return ds


def read_gml_co2(filename=None, convert_missing=True, **kwargs):
    """
    Function to read carbon dioxide data from NOAA GML.

    Parameters
    ----------
    filename : str or pathlib.Path
        Data file full path name.
    convert_missing : boolean
        Option to convert missing values to NaN. If turned off will
        set variable attribute to missing value expected. This works well
        to preserve the data type best for writing to a netCDF file.

    Returns
    -------
    dataset : Xarray.dataset
        Standard ARM Xarray dataset with the data cleaned up to have units,
        long_name, correct type and some other stuff.
    **kwargs : keywords
        Keywords to pass through to ACT read_csv() routine.

    """
    ds = None
    if filename is None:
        return ds

    variables = {
        'site_code': None,
        'year': None,
        'month': None,
        'day': None,
        'hour': None,
        'minute': None,
        'second': None,
        'time_decimal': None,
        'value': {
            'long_name': 'Carbon monoxide in dry air',
            'units': 'ppm',
            '_FillValue': -999.99,
            'comment': (
                'Mole fraction reported in units of micromol mol-1 '
                '(10-6 mol per mol of dry air); abbreviated as ppm (parts per million).'
            ),
            '__type': np.float32,
            '__rename': 'co2',
        },
        'value_std_dev': {
            'long_name': 'Carbon monoxide in dry air',
            'units': 'ppm',
            '_FillValue': -99.99,
            'comment': (
                'This is the standard deviation of the reported mean value '
                'when nvalue is greater than 1. See provider_comment if available.'
            ),
            '__type': np.float32,
            '__rename': 'co2_std_dev',
        },
        'nvalue': {
            'long_name': 'Number of measurements contributing to reported value',
            'units': '1',
            '_FillValue': -9,
            '__type': np.int16,
            '__rename': 'number_of_measurements',
        },
        'latitude': {
            'long_name': 'Latitude at which air sample was collected',
            'units': 'degrees_north',
            '_FillValue': -999.999,
            'standard_name': 'latitude',
            '__type': np.float32,
        },
        'longitude': {
            'long_name': 'Latitude at which air sample was collected',
            'units': 'degrees_east',
            '_FillValue': -999.999,
            'standard_name': 'longitude',
            '__type': np.float32,
        },
        'altitude': {
            'long_name': 'Sample altitude',
            'units': 'm',
            '_FillValue': -999.999,
            'standard_name': 'altitude',
            'comment': (
                'Altitude for this dataset is the sum of surface elevation '
                '(masl) and sample intake height (magl)'
            ),
            '__type': np.float32,
        },
        'intake_height': {
            'long_name': 'Sample intake height above ground level',
            'units': 'm',
            '_FillValue': -999.999,
            '__type': np.float32,
        },
    }

    test_filename = filename
    if isinstance(test_filename, (list, tuple)):
        test_filename = test_filename[0]

    with open(test_filename) as fc:
        skiprows = int(fc.readline().strip().split()[-1]) - 1

    ds = act.io.csvfiles.read_csv(filename, sep=r'\s+', skiprows=skiprows, **kwargs)

    timestamp = np.full(ds['year'].size, np.nan, dtype='datetime64[s]')
    for ii in range(0, len(timestamp)):
        ts = datetime(
            ds['year'].values[ii],
            ds['month'].values[ii],
            ds['day'].values[ii],
            ds['hour'].values[ii],
            ds['minute'].values[ii],
            ds['second'].values[ii],
        )
        timestamp[ii] = np.datetime64(ts)

    ds = ds.rename({'index': 'time'})
    ds = ds.assign_coords(time=timestamp)
    ds['time'].attrs['long_name'] = 'Time'

    for var_name, value in variables.items():
        if value is None:
            del ds[var_name]
        else:
            for att_name, att_value in value.items():
                if att_name == '__type':
                    values = ds[var_name].values
                    values = values.astype(att_value)
                    ds[var_name].values = values
                elif att_name == '__rename':
                    ds = ds.rename({var_name: att_value})
                else:
                    ds[var_name].attrs[att_name] = att_value

            if convert_missing:
                try:
                    var_name = variables[var_name]['__rename']
                except KeyError:
                    pass

                try:
                    missing_value = ds[var_name].attrs['_FillValue']
                    values = ds[var_name].values.astype(float)
                    values[np.isclose(missing_value, values)] = np.nan
                    ds[var_name].values = values
                    ds[var_name].attrs['_FillValue'] = np.nan
                except KeyError:
                    pass

    values = ds['qcflag'].values
    bad_index = []
    suspect_index = []
    for ii, value in enumerate(values):
        pts = list(value)
        if pts[0] != '.':
            bad_index.append(ii)
        if pts[1] != '.':
            suspect_index.append(ii)

    var_name = 'co2'
    qc_var_name = ds.qcfilter.create_qc_variable(var_name)
    ds.qcfilter.add_test(
        var_name,
        index=bad_index,
        test_assessment='Bad',
        test_meaning='Obvious problems during collection or analysis',
    )
    ds.qcfilter.add_test(
        var_name,
        index=suspect_index,
        test_assessment='Indeterminate',
        test_meaning=(
            'Likely valid but does not meet selection criteria determined by '
            'the goals of a particular investigation'
        ),
    )
    ds[qc_var_name].attrs[
        'comment'
    ] = 'This quality control flag is provided by the contributing PIs'
    del ds['qcflag']

    return ds


def read_gml_ozone(filename=None, **kwargs):
    """
    Function to read carbon dioxide data from NOAA GML.

    Parameters
    ----------
    filename : str or pathlib.Path
        Data file full path name.
    **kwargs : keywords
        Keywords to pass through to ACT read_csv() routine.

    Returns
    -------
    dataset : Xarray.dataset
        Standard ARM Xarray dataset with the data cleaned up to have units,
        long_name, correct type and some other stuff.
    """

    ds = None
    if filename is None:
        return ds

    test_filename = filename
    if isinstance(test_filename, (list, tuple)):
        test_filename = test_filename[0]

    with open(test_filename) as fc:
        skiprows = 0
        while True:
            line = fc.readline().strip().split()
            try:
                if len(line) == 6 and line[0] == 'STN':
                    break
            except IndexError:
                pass
            skiprows += 1

    ds = act.io.csvfiles.read_csv(filename, sep=r'\s+', skiprows=skiprows, **kwargs)
    ds.attrs['station'] = str(ds['STN'].values[0]).lower()

    timestamp = np.full(ds['YEAR'].size, np.nan, dtype='datetime64[s]')
    for ii in range(0, len(timestamp)):
        ts = datetime(
            ds['YEAR'].values[ii],
            ds['MON'].values[ii],
            ds['DAY'].values[ii],
            ds['HR'].values[ii],
        )
        timestamp[ii] = np.datetime64(ts)

    ds = ds.rename({'index': 'time'})
    ds = ds.assign_coords(time=timestamp)
    ds['time'].attrs['long_name'] = 'Time'

    for var_name in ['STN', 'YEAR', 'MON', 'DAY', 'HR']:
        del ds[var_name]

    var_name = 'ozone'
    ds = ds.rename({'O3(PPB)': var_name})
    ds[var_name].attrs['long_name'] = 'Ozone'
    ds[var_name].attrs['units'] = 'ppb'
    ds[var_name].attrs['_FillValue'] = np.nan
    ds[var_name].values = ds[var_name].values.astype(np.float32)

    return ds


def read_gml_radiation(filename=None, convert_missing=True, remove_time_vars=True, **kwargs):
    """
    Function to read radiation data from NOAA GML.

    Parameters
    ----------
    filename : str or pathlib.Path
        Data file full path name.
    convert_missing : boolean
        Option to convert missing values to NaN. If turned off will
        set variable attribute to missing value expected. This works well
        to preserve the data type best for writing to a netCDF file.
    remove_time_vars : boolean
        Some column names in the CSV file are used for creating the time
        coordinate variable in the returend Xarray DataSet. Once used the
        variables are not needed and will be removed from DataSet.
    **kwargs : keywords
        Keywords to pass through to ACT read_csv() routine.

    Returns
    -------
    dataset : Xarray.dataset
        Standard ARM Xarray dataset with the data cleaned up to have units,
        long_name, correct type and some other stuff.
    """

    ds = None
    if filename is None:
        return ds

    column_names = {
        'year': None,
        'jday': None,
        'month': None,
        'day': None,
        'hour': None,
        'minute': None,
        'decimal_time': None,
        'solar_zenith_angle': {
            'units': 'degree',
            'long_name': 'Solar zenith angle',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'downwelling_global_solar': {
            'units': 'W/m^2',
            'long_name': 'Downwelling global solar',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'upwelling_global_solar': {
            'units': 'W/m^2',
            'long_name': 'Upwelling global solar',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'direct_normal_solar': {
            'units': 'W/m^2',
            'long_name': 'Direct-normal solar',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'downwelling_diffuse_solar': {
            'units': 'W/m^2',
            'long_name': 'Downwelling diffuse solar',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'downwelling_thermal_infrared': {
            'units': 'W/m^2',
            'long_name': 'Downwelling thermal infrared',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'downwelling_infrared_case_temp': {
            'units': 'degK',
            'long_name': 'Downwelling infrared case temp',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'downwelling_infrared_dome_temp': {
            'units': 'degK',
            'long_name': 'downwelling infrared dome temp',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'upwelling_thermal_infrared': {
            'units': 'W/m^2',
            'long_name': 'Upwelling thermal infrared',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'upwelling_infrared_case_temp': {
            'units': 'degK',
            'long_name': 'Upwelling infrared case temp',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'upwelling_infrared_dome_temp': {
            'units': 'degK',
            'long_name': 'Upwelling infrared dome temp',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'global_UVB': {
            'units': 'mW/m^2',
            'long_name': 'global ultraviolet-B',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'par': {
            'units': 'W/m^2',
            'long_name': 'Photosynthetically active radiation',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'net_solar': {
            'units': 'W/m^2',
            'long_name': 'Net solar (downwelling_global_solar - upwelling_global_solar)',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'net_infrared': {
            'units': 'W/m^2',
            'long_name': (
                'Net infrared (downwelling_thermal_infrared - ' 'upwelling_thermal_infrared)'
            ),
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'net_radiation': {
            'units': 'W/m^2',
            'long_name': 'Net radiation (net_solar + net_infrared)',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'air_temperature_10m': {
            'units': 'degC',
            'long_name': '10-meter air temperature',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'relative_humidity': {
            'units': '%',
            'long_name': 'Relative humidity',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'wind_speed': {
            'units': 'm/s',
            'long_name': 'Wind speed',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'wind_direction': {
            'units': 'degree',
            'long_name': 'Wind direction (clockwise from north)',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
        'station_pressure': {
            'units': 'millibar',
            'long_name': 'Station atmospheric pressure',
            '_FillValue': -9999.9,
            '__type': np.float32,
        },
    }

    # Add additinal column names for NOAA SPASH campaign
    if str(Path(filename).name).startswith('cbc') or str(Path(filename).name).startswith('ckp'):
        column_names['SPN1_total'] = {
            'units': 'W/m^2',
            'long_name': 'SPN1 total average',
            '_FillValue': -9999.9,
            '__type': np.float32
        }
        column_names['SPN1_diffuse'] = {
            'units': 'W/m^2',
            'long_name': 'SPN1 diffuse average',
            '_FillValue': -9999.9,
            '__type': np.float32
        }

    names = list(column_names.keys())
    skip_vars = [
        'year',
        'jday',
        'month',
        'day',
        'hour',
        'minute',
        'decimal_time',
        'solar_zenith_angle',
    ]
    num = 1
    for ii, name in enumerate(column_names.keys()):
        if name in skip_vars:
            continue
        names.insert(ii + num, 'qc_' + name)
        num += 1

    ds = act.io.csvfiles.read_csv(filename, sep=r'\s+', header=None, skiprows=2, column_names=names, **kwargs)

    if isinstance(filename, (list, tuple)):
        filename = filename[0]

    if ds is not None:
        with open(filename) as fc:
            lat = None
            lon = None
            alt = None
            alt_unit = None
            station = None
            for ii in [0, 1]:
                line = fc.readline().strip().split()
                if len(line) == 1:
                    station = line[0]
                else:
                    lat = np.array(line[0], dtype=np.float32)
                    lon = np.array(line[1], dtype=np.float32)
                    alt = np.array(line[2], dtype=np.float32)
                    alt_unit = str(line[3])

        ds['lat'] = xr.DataArray(
            lat,
            attrs={
                'long_name': 'Latitude',
                'units': 'degree_north',
                'standard_name': 'latitude',
            },
        )
        ds['lon'] = xr.DataArray(
            lon,
            attrs={
                'long_name': 'Longitude',
                'units': 'degree_east',
                'standard_name': 'longitude',
            },
        )
        ds['alt'] = xr.DataArray(
            alt,
            attrs={
                'long_name': 'Latitude',
                'units': alt_unit,
                'standard_name': 'altitude',
            },
        )
        ds.attrs['location'] = station

        timestamp = np.full(ds['year'].size, np.nan, dtype='datetime64[s]')
        for ii in range(0, len(timestamp)):
            ts = datetime(
                ds['year'].values[ii],
                ds['month'].values[ii],
                ds['day'].values[ii],
                ds['hour'].values[ii],
                ds['minute'].values[ii],
            )
            timestamp[ii] = np.datetime64(ts)

        ds = ds.rename({'index': 'time'})
        ds = ds.assign_coords(time=timestamp)
        ds['time'].attrs['long_name'] = 'Time'
        for var_name, value in column_names.items():
            if value is None:
                ds[var_name]
            else:
                for att_name, att_value in value.items():
                    if att_name == '__type':
                        values = ds[var_name].values
                        values = values.astype(att_value)
                        ds[var_name].values = values
                    else:
                        ds[var_name].attrs[att_name] = att_value

                if convert_missing:
                    try:
                        missing_value = ds[var_name].attrs['_FillValue']
                        values = ds[var_name].values.astype(float)
                        index = np.isclose(values, missing_value)
                        values[index] = np.nan
                        ds[var_name].values = values
                        ds[var_name].attrs['_FillValue'] = np.nan
                    except KeyError:
                        pass

        for var_name in ds.data_vars:
            if not var_name.startswith('qc_'):
                continue
            data_var_name = var_name.replace('qc_', '', 1)
            attrs = {
                'long_name': f"Quality control variable for: {ds[data_var_name].attrs['long_name']}",
                'units': '1',
                'standard_name': 'quality_flag',
                'flag_values': [0, 1, 2],
                'flag_meanings': [
                    'Not failing any tests',
                    'Knowingly bad value',
                    'Should be used with scrutiny',
                ],
                'flag_assessments': ['Good', 'Bad', 'Indeterminate'],
            }
            ds[var_name].attrs = attrs
            ds[data_var_name].attrs['ancillary_variables'] = var_name

        if remove_time_vars:
            remove_var_names = [
                'year',
                'jday',
                'month',
                'day',
                'hour',
                'minute',
                'decimal_time',
            ]
            ds = ds.drop_vars(remove_var_names)

    return ds


def read_gml_met(filename=None, convert_missing=True, **kwargs):
    """
    Function to read meteorlogical data from NOAA GML.

    Parameters
    ----------
    filename : str or pathlib.Path
        Data file full path name.
    convert_missing : boolean
        Option to convert missing values to NaN. If turned off will
        set variable attribute to missing value expected. This works well
        to preserve the data type best for writing to a netCDF file.
    **kwargs : keywords
        Keywords to pass through to ACT read_csv() routine.

    Returns
    -------
    dataset : Xarray.dataset
        Standard ARM Xarray dataset with the data cleaned up to have units,
        long_name, correct type and some other stuff.
    """

    ds = None
    if filename is None:
        return ds

    column_names = {
        'station': None,
        'year': None,
        'month': None,
        'day': None,
        'hour': None,
        'minute': None,
        'wind_direction': {
            'units': 'degree',
            'long_name': 'Average wind direction from which the wind is blowing',
            '_FillValue': -999,
            '__type': np.int16,
        },
        'wind_speed': {
            'units': 'm/s',
            'long_name': 'Average wind speed',
            '_FillValue': -999.9,
            '__type': np.float32,
        },
        'wind_steadiness_factor': {
            'units': '1',
            'long_name': '100 times the ratio of the vector wind speed to the '
            'average wind speed for the hour',
            '_FillValue': -9,
            '__type': np.int16,
        },
        'barometric_pressure': {
            'units': 'hPa',
            'long_name': 'Station barometric pressure',
            '_FillValue': -999.90,
            '__type': np.float32,
        },
        'temperature_2m': {
            'units': 'degC',
            'long_name': 'Temperature at 2 meters above ground level',
            '_FillValue': -999.9,
            '__type': np.float32,
        },
        'temperature_10m': {
            'units': 'degC',
            'long_name': 'Temperature at 10 meters above ground level',
            '_FillValue': -999.9,
            '__type': np.float32,
        },
        'temperature_tower_top': {
            'units': 'degC',
            'long_name': 'Temperature at top of instrument tower',
            '_FillValue': -999.9,
            '__type': np.float32,
        },
        'realitive_humidity': {
            'units': 'percent',
            'long_name': 'Relative humidity',
            '_FillValue': -99,
            '__type': np.int16,
        },
        'preciptation_intensity': {
            'units': 'mm/hour',
            'long_name': 'Amount of precipitation per hour',
            '_FillValue': -99,
            '__type': np.int16,
            'comment': (
                'The precipitation amount is measured with an unheated '
                'tipping bucket rain gauge.'
            ),
        },
    }

    minutes = True
    test_filename = filename
    if isinstance(test_filename, (list, tuple)):
        test_filename = test_filename[0]

    if '_hour_' in Path(test_filename).name:
        minutes = False
        del column_names['minute']

    ds = act.io.csvfiles.read_csv(filename, sep=r'\s+', header=None, column_names=column_names.keys(), **kwargs)

    if ds is not None:

        timestamp = np.full(ds['year'].size, np.nan, dtype='datetime64[s]')
        for ii in range(0, len(timestamp)):
            if minutes:
                ts = datetime(
                    ds['year'].values[ii],
                    ds['month'].values[ii],
                    ds['day'].values[ii],
                    ds['hour'].values[ii],
                    ds['minute'].values[ii],
                )
            else:
                ts = datetime(
                    ds['year'].values[ii],
                    ds['month'].values[ii],
                    ds['day'].values[ii],
                    ds['hour'].values[ii],
                )

            timestamp[ii] = np.datetime64(ts)

        ds = ds.rename({'index': 'time'})
        ds = ds.assign_coords(time=timestamp)
        ds['time'].attrs['long_name'] = 'Time'
        for var_name, value in column_names.items():

            if value is None:
                del ds[var_name]
            else:
                for att_name, att_value in value.items():
                    if att_name == '__type':
                        values = ds[var_name].values
                        values = values.astype(att_value)
                        ds[var_name].values = values
                    else:
                        ds[var_name].attrs[att_name] = att_value

                if convert_missing:
                    try:
                        missing_value = ds[var_name].attrs['_FillValue']
                        values = ds[var_name].values.astype(float)
                        index = np.isclose(values, missing_value)
                        values[index] = np.nan
                        ds[var_name].values = values
                        ds[var_name].attrs['_FillValue'] = np.nan
                    except KeyError:
                        pass

    return ds



"""
ACT: The Atmospheric Community Toolkit
======================================

"""

# No more pandas warnings
from pandas.plotting import register_matplotlib_converters

from . import corrections, discovery, io, plotting, qc, retrievals, tests, utils
from ._version import get_versions

register_matplotlib_converters()

# Version for source builds
vdict = get_versions()
__version__ = vdict['version']



"""
This module contains various retrievals for datasets.

"""

from .aeri import aeri2irt
from .cbh import generic_sobel_cbh
from .doppler_lidar import compute_winds_from_ppi
from .irt import sst_from_irt, sum_function_irt
from .radiation import (
    calculate_dsh_from_dsdh_sdn,
    calculate_irradiance_stats,
    calculate_longwave_radiation,
    calculate_net_radiation,
)
from .sonde import (
    calculate_pbl_liu_liang,
    calculate_precipitable_water,
    calculate_stability_indicies,
)
from .sp2 import calc_sp2_diams_masses, process_sp2_psds



"""
This module contains sample files used for testing the ARM Community Toolkit.
Files in this module should only be used for testing, not production.

"""

from .sample_files import (
    EXAMPLE_AERI,
    EXAMPLE_ANL_CSV,
    EXAMPLE_AOSMET,
    EXAMPLE_BRS,
    EXAMPLE_CEIL1,
    EXAMPLE_CEIL_WILDCARD,
    EXAMPLE_CO2FLX4M,
    EXAMPLE_DLPPI,
    EXAMPLE_EBBR1,
    EXAMPLE_EBBR2,
    EXAMPLE_IRTSST,
    EXAMPLE_LCL1,
    EXAMPLE_MET1,
    EXAMPLE_MET_CONTOUR,
    EXAMPLE_MET_CSV,
    EXAMPLE_MET_TEST1,
    EXAMPLE_MET_TEST2,
    EXAMPLE_MET_WILDCARD,
    EXAMPLE_METE40,
    EXAMPLE_MFRSR,
    EXAMPLE_MPL_1SAMPLE,
    EXAMPLE_NAV,
    EXAMPLE_NOAA_PSL,
    EXAMPLE_RL1,
    EXAMPLE_SIGMA_MPLV5,
    EXAMPLE_SIRS,
    EXAMPLE_SONDE1,
    EXAMPLE_SONDE_WILDCARD,
    EXAMPLE_STAMP_WILDCARD,
    EXAMPLE_SURFSPECALB1MLAWER,
    EXAMPLE_TWP_SONDE_20060121,
    EXAMPLE_IRT25m20s,
    EXAMPLE_HK,
    EXAMPLE_INI,
    EXAMPLE_SP2B
)



"""
Modules for reading in NOAA PSL data.
"""

import datetime as dt

import numpy as np
import pandas as pd
import xarray as xr


def read_psl_wind_profiler(filename, transpose=True):
    """
    Returns `xarray.Dataset` with stored data and metadata from a user-defined
    NOAA PSL wind profiler file.

    Parameters
    ----------
    filename : str
        Name of file(s) to read.
    transpose : bool
        True to transpose the data.

    Return
    ------
    obj_low :  Xarray.dataset
        Standard Xarray dataset with the data for low mode
    obj_high : Xarray.dataset
        Standard Xarray dataset with the data for high mode.

    """
    # read file with pandas for preparation.
    df = pd.read_csv(filename, header=None)

    # Get location of where each table begins
    index_list = df[0] == ' CTD'
    idx = np.where(index_list)

    # Get header of each column of data.
    column_list = list(df.loc[9][0].split())

    beam_vars = ['RAD', 'CNT', 'SNR', 'QC']
    for i, c in enumerate(column_list):
        if c in beam_vars:
            if column_list.count(c) > 2:
                column_list[i] = c + '1'
            elif column_list.count(c) > 1:
                column_list[i] = c + '2'
            elif column_list.count(c) > 0:
                column_list[i] = c + '3'

    # Loop through column data only which appears after 10 lines of metadata.
    # Year, Month, day, hour, minute, second, utc offset
    low = []
    hi = []
    for i in range(idx[0].shape[0] - 1):
        # index each table by using the idx of when CTD appears.
        # str split is use as 2 spaces are added to each data point,
        # convert to float.
        date_str = df.iloc[idx[0][i] + 3]
        date_str = list(filter(None, date_str[0].split(' ')))
        date_str = list(map(int, date_str))
        # Datetime not taking into account the utc offset yet
        time = dt.datetime(
            2000 + date_str[0],
            date_str[1],
            date_str[2],
            date_str[3],
            date_str[4],
            date_str[5],
        )

        mode = df.iloc[idx[0][i] + 7][0]
        mode = int(mode.split(' ')[-1])

        df_array = np.array(
            df.iloc[idx[0][i] + 10 : idx[0][i + 1] - 1][0].str.split(r'\s{2,}').tolist(),
            dtype='float',
        )
        df_add = pd.DataFrame(df_array, columns=column_list)
        df_add = df_add.replace(999999.0, np.nan)

        xr_add = df_add.to_xarray()
        xr_add = xr_add.swap_dims({'index': 'height'})
        xr_add = xr_add.reset_coords('index')
        xr_add = xr_add.assign_coords({'time': np.array(time), 'height': xr_add['HT'].values})

        if mode < 1000.0:
            low.append(xr_add)
        else:
            hi.append(xr_add)

    obj_low = xr.concat(low, 'time')
    obj_hi = xr.concat(hi, 'time')

    # Adding site information line 1
    site_loc = df.iloc[idx[0][0]]
    site_list = site_loc.str.split(r'\s{2}').tolist()
    site = site_list[0][0].strip()

    obj_low.attrs['site_identifier'] = site
    obj_hi.attrs['site_identifier'] = site

    # Adding data type and revision number line 2.
    rev = df.loc[idx[0][0] + 1]
    rev_list = rev.str.split(r'\s{3}').tolist()
    rev_array = np.array(rev_list[0])

    obj_low.attrs['data_type'] = rev_array[0].strip()
    obj_hi.attrs['data_type'] = rev_array[0].strip()
    obj_low.attrs['revision_number'] = rev_array[1].strip()
    obj_hi.attrs['revision_number'] = rev_array[1].strip()

    # Adding coordinate attributes line 3.
    coords = df.loc[idx[0][0] + 2]
    coords_list = coords.str.split(r'\s{2,}').tolist()
    coords_list[0].remove('')
    coords_array = np.array(coords_list[0], dtype='float32')

    obj_low.attrs['latitude'] = np.array([coords_array[0]])
    obj_hi.attrs['latitude'] = np.array([coords_array[0]])
    obj_low.attrs['longitude'] = np.array([coords_array[1]])
    obj_hi.attrs['longitude'] = np.array([coords_array[1]])
    obj_low.attrs['altitude'] = np.array([coords_array[2]])
    obj_hi.attrs['altitude'] = np.array([coords_array[2]])

    # Adding azimuth and elevation line 9
    az_el = df.loc[idx[0][0] + 8]
    az_el_list = az_el.str.split(r'\s{2,}').tolist()
    az_el_list[0].remove('')
    az_el_array = np.array(az_el_list[0])
    az = []
    el = []
    for i in az_el_array:
        sep = i.split()
        az.append(sep[0])
        el.append(sep[1])
    az_array = np.array(az, dtype='float32')
    el_array = np.array(el, dtype='float32')

    obj_low.attrs['azimuth'] = az_array
    obj_hi.attrs['azimuth'] = az_array
    obj_low.attrs['elevation'] = el_array
    obj_hi.attrs['elevation'] = el_array

    if transpose:
        obj_low = obj_low.transpose()
        obj_hi = obj_hi.transpose()

    return obj_low, obj_hi



"""
Script for downloading ASOS data from the Iowa Mesonet API

"""

import json
import time
import warnings
from datetime import datetime

import numpy as np
import pandas as pd
import xarray as xr
from six import StringIO

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen


def get_asos(time_window, lat_range=None, lon_range=None, station=None):
    """
    Returns all of the station observations from the Iowa Mesonet from either
    a given latitude and longitude window or a given station code.

    Parameters
    ----------
    time_window: tuple
        A 2 member list or tuple containing the start and end times. The
        times must be python datetimes.
    lat_range: tuple
        The latitude window to grab all of the ASOS observations from.
    lon_range: tuple
        The longitude window to grab all of the ASOS observations from.
    station: str
        The station ID to grab the ASOS observations from.

    Returns
    -------
    asos_ds: dict of xarray datasets
        A dictionary of ACT datasets whose keys are the ASOS station IDs.

    Examples
    --------
    If you want to obtain timeseries of ASOS observations for Chicago O'Hare
    Airport, simply do::

        $ time_window = [datetime(2020, 2, 4, 2, 0), datetime(2020, 2, 10, 10, 0)]
        $ station = "KORD"
        $ my_asoses = act.discovery.get_asos(time_window, station="ORD")
    """

    # First query the database for all of the JSON info for every station
    # Only add stations whose lat/lon are within the Grid's boundaries
    regions = """AF AL_ AI_ AQ_ AG_ AR_ AK AL AM_
        AO_ AS_ AR AW_ AU_ AT_
        AZ_ BA_ BE_ BB_ BG_ BO_ BR_ BF_
        BT_ BS_ BI_ BM_ BB_ BY_ BZ_ BJ_ BW_ AZ CA CA_AB
        CA_BC CD_ CK_ CF_ CG_ CL_ CM_ CO CO_ CN_ CR_ CT
        CU_ CV_ CY_ CZ_ DE DK_ DJ_ DM_ DO_
        DZ EE_ ET_ FK_ FM_ FJ_ FI_ FR_ GF_ PF_
        GA_ GM_ GE_ DE_ GH_ GI_ KY_ GB_ GR_ GL_ GD_
        GU_ GT_ GN_ GW_ GY_ HT_ HN_ HK_ HU_ IS_ IN_
        ID_ IR_ IQ_ IE_ IL_ IT_ CI_ JM_ JP_
        JO_ KZ_ KE_ KI_ KW_ LA_ LV_ LB_ LS_ LR_ LY_
        LT_ LU_ MK_ MG_ MW_ MY_ MV_ ML_ CA_MB
        MH_ MR_ MU_ YT_ MX_ MD_ MC_ MA_ MZ_ MM_ NA_ NP_
        AN_ NL_ CA_NB NC_ CA_NF NF_ NI_
        NE_ NG_ MP_ KP_ CA_NT NO_ CA_NS CA_NU OM_
        CA_ON PK_ PA_ PG_ PY_ PE_ PH_ PN_ PL_
        PT_ CA_PE PR_ QA_ CA_QC RO_ RU_RW_ SH_ KN_
        LC_ VC_ WS_ ST_ CA_SK SA_ SN_ RS_ SC_
        SL_ SG_ SK_ SI_ SB_ SO_ ZA_ KR_ ES_ LK_ SD_ SR_
        SZ_ SE_ CH_ SY_ TW_ TJ_ TZ_ TH_
        TG_ TO_ TT_ TU TN_ TR_ TM_ UG_ UA_ AE_ UN_ UY_
        UZ_ VU_ VE_ VN_ VI_ YE_ CA_YT ZM_ ZW_
        EC_ EG_ FL GA GQ_ HI HR_ IA ID IL IO_ IN KS
        KH_ KY KM_ LA MA MD ME
        MI MN MO MS MT NC ND NE NH NJ NM NV NY OH OK
        OR PA RI SC SV_ SD TD_ TN TX UT VA VT VG_
        WA WI WV WY"""

    networks = ['AWOS']
    metadata_list = {}
    if lat_range is not None and lon_range is not None:
        lon_min, lon_max = lon_range
        lat_min, lat_max = lat_range
        for region in regions.split():
            networks.append(f'{region}_ASOS')

        site_list = []
        for network in networks:
            # Get metadata
            uri = ('https://mesonet.agron.iastate.edu/' 'geojson/network/%s.geojson') % (network,)
            data = urlopen(uri)
            jdict = json.load(data)
            for site in jdict['features']:
                lat = site['geometry']['coordinates'][1]
                lon = site['geometry']['coordinates'][0]
                if lat >= lat_min and lat <= lat_max:
                    if lon >= lon_min and lon <= lon_max:
                        station_metadata_dict = {}
                        station_metadata_dict['site_latitude'] = lat
                        station_metadata_dict['site_longitude'] = lat
                        for my_keys in site['properties']:
                            station_metadata_dict[my_keys] = site['properties'][my_keys]
                        metadata_list[site['properties']['sid']] = station_metadata_dict
                        site_list.append(site['properties']['sid'])
    elif station is not None:
        site_list = [station]
        for region in regions.split():
            networks.append(f'{region}_ASOS')
        for network in networks:
            # Get metadata
            uri = ('https://mesonet.agron.iastate.edu/' 'geojson/network/%s.geojson') % (network,)
            data = urlopen(uri)
            jdict = json.load(data)
            for site in jdict['features']:
                lat = site['geometry']['coordinates'][1]
                lon = site['geometry']['coordinates'][0]
                if site['properties']['sid'] == station:
                    station_metadata_dict = {}
                    station_metadata_dict['site_latitude'] = lat
                    station_metadata_dict['site_longitude'] = lon
                    for my_keys in site['properties']:
                        if my_keys == 'elevation':
                            station_metadata_dict['elevation'] = (
                                '%f meter' % site['properties'][my_keys]
                            )
                        else:
                            station_metadata_dict[my_keys] = site['properties'][my_keys]
                    metadata_list[station] = station_metadata_dict

        # Get station metadata
    else:
        raise ValueError('Either both lat_range and lon_range or station must ' + 'be specified!')

    # Get the timestamp for each request
    start_time = time_window[0]
    end_time = time_window[1]

    SERVICE = 'http://mesonet.agron.iastate.edu/cgi-bin/request/asos.py?'
    service = SERVICE + 'data=all&tz=Etc/UTC&format=comma&latlon=yes&'

    service += start_time.strftime('year1=%Y&month1=%m&day1=%d&hour1=%H&minute1=%M&')
    service += end_time.strftime('year2=%Y&month2=%m&day2=%d&hour2=%H&minute2=%M')
    station_obs = {}
    for stations in site_list:
        uri = f'{service}&station={stations}'
        print(f'Downloading: {stations}')
        data = _download_data(uri)
        buf = StringIO()
        buf.write(data)
        buf.seek(0)

        my_df = pd.read_csv(buf, skiprows=5, na_values='M')

        if len(my_df['lat'].values) == 0:
            warnings.warn(
                'No data available at station %s between time %s and %s'
                % (
                    stations,
                    start_time.strftime('%Y-%m-%d %H:%M:%S'),
                    end_time.strftime('%Y-%m-%d %H:%M:%S'),
                )
            )
        else:

            def to_datetime(x):
                return datetime.strptime(x, '%Y-%m-%d %H:%M')

            my_df['time'] = my_df['valid'].apply(to_datetime)
            my_df = my_df.set_index('time')
            my_df = my_df.drop('valid', axis=1)
            my_df = my_df.drop('station', axis=1)
            my_df = my_df.to_xarray()

            my_df.attrs = metadata_list[stations]
            my_df['lon'].attrs['units'] = 'degree'
            my_df['lon'].attrs['long_name'] = 'Longitude'
            my_df['lat'].attrs['units'] = 'degree'
            my_df['lat'].attrs['long_name'] = 'Latitude'

            my_df['tmpf'].attrs['units'] = 'degrees Fahrenheit'
            my_df['tmpf'].attrs['long_name'] = 'Temperature in degrees Fahrenheit'

            # Fahrenheit to Celsius
            my_df['temp'] = (5.0 / 9.0 * my_df['tmpf']) - 32.0
            my_df['temp'].attrs['units'] = 'degrees Celsius'
            my_df['temp'].attrs['long_name'] = 'Temperature in degrees Celsius'
            my_df['dwpf'].attrs['units'] = 'degrees Fahrenheit'
            my_df['dwpf'].attrs['long_name'] = 'Dewpoint temperature in degrees Fahrenheit'

            # Fahrenheit to Celsius
            my_df['dwpc'] = (5.0 / 9.0 * my_df['tmpf']) - 32.0
            my_df['dwpc'].attrs['units'] = 'degrees Celsius'
            my_df['dwpc'].attrs['long_name'] = 'Dewpoint temperature in degrees Celsius'
            my_df['relh'].attrs['units'] = 'percent'
            my_df['relh'].attrs['long_name'] = 'Relative humidity'
            my_df['drct'].attrs['units'] = 'degrees'
            my_df['drct'].attrs['long_name'] = 'Wind speed in degrees'
            my_df['sknt'].attrs['units'] = 'knots'
            my_df['sknt'].attrs['long_name'] = 'Wind speed in knots'
            my_df['spdms'] = my_df['sknt'] * 0.514444
            my_df['spdms'].attrs['units'] = 'm s-1'
            my_df['spdms'].attrs['long_name'] = 'Wind speed in meters per second'
            my_df['u'] = -np.sin(np.deg2rad(my_df['drct'])) * my_df['spdms']
            my_df['u'].attrs['units'] = 'm s-1'
            my_df['u'].attrs['long_name'] = 'Zonal component of surface wind'
            my_df['v'] = -np.cos(np.deg2rad(my_df['drct'])) * my_df['spdms']
            my_df['v'].attrs['units'] = 'm s-1'
            my_df['v'].attrs['long_name'] = 'Meridional component of surface wind'
            my_df['mslp'].attrs['units'] = 'mb'
            my_df['mslp'].attrs['long_name'] = 'Mean Sea Level Pressure'
            my_df['alti'].attrs['units'] = 'in Hg'
            my_df['alti'].attrs['long_name'] = 'Atmospheric pressure in inches of Mercury'
            my_df['vsby'].attrs['units'] = 'mi'
            my_df['vsby'].attrs['long_name'] = 'Visibility'
            my_df['vsbykm'] = my_df['vsby'] * 1.60934
            my_df['vsbykm'].attrs['units'] = 'km'
            my_df['vsbykm'].attrs['long_name'] = 'Visibility'
            my_df['gust'] = my_df['gust'] * 0.514444
            my_df['gust'].attrs['units'] = 'm s-1'
            my_df['gust'].attrs['long_name'] = 'Wind gust speed'
            my_df['skyc1'].attrs['long_name'] = 'Sky level 1 coverage'
            my_df['skyc2'].attrs['long_name'] = 'Sky level 2 coverage'
            my_df['skyc3'].attrs['long_name'] = 'Sky level 3 coverage'
            my_df['skyc4'].attrs['long_name'] = 'Sky level 4 coverage'
            my_df['skyl1'] = my_df['skyl1'] * 0.3048
            my_df['skyl2'] = my_df['skyl2'] * 0.3048
            my_df['skyl3'] = my_df['skyl3'] * 0.3048
            my_df['skyl4'] = my_df['skyl4'] * 0.3048
            my_df['skyl1'].attrs['long_name'] = 'Sky level 1 altitude'
            my_df['skyl2'].attrs['long_name'] = 'Sky level 2 altitude'
            my_df['skyl3'].attrs['long_name'] = 'Sky level 3 altitude'
            my_df['skyl4'].attrs['long_name'] = 'Sky level 4 altitude'
            my_df['skyl1'].attrs['long_name'] = 'meter'
            my_df['skyl2'].attrs['long_name'] = 'meter'
            my_df['skyl3'].attrs['long_name'] = 'meter'
            my_df['skyl4'].attrs['long_name'] = 'meter'

            my_df['wxcodes'].attrs['long_name'] = 'Weather code'
            my_df['ice_accretion_1hr'] = my_df['ice_accretion_1hr'] * 2.54
            my_df['ice_accretion_1hr'].attrs['units'] = 'cm'
            my_df['ice_accretion_1hr'].attrs['long_name'] = '1 hour ice accretion'
            my_df['ice_accretion_3hr'] = my_df['ice_accretion_3hr'] * 2.54
            my_df['ice_accretion_3hr'].attrs['units'] = 'cm'
            my_df['ice_accretion_3hr'].attrs['long_name'] = '3 hour ice accretion'
            my_df['ice_accretion_6hr'] = my_df['ice_accretion_3hr'] * 2.54
            my_df['ice_accretion_6hr'].attrs['units'] = 'cm'
            my_df['ice_accretion_6hr'].attrs['long_name'] = '6 hour ice accretion'
            my_df['peak_wind_gust'] = my_df['peak_wind_gust'] * 0.514444
            my_df['peak_wind_gust'].attrs['units'] = 'm s-1'
            my_df['peak_wind_gust'].attrs['long_name'] = 'Peak wind gust speed'
            my_df['peak_wind_drct'].attrs['drct'] = 'degree'
            my_df['peak_wind_drct'].attrs['long_name'] = 'Peak wind gust direction'
            my_df['u_peak'] = -np.sin(np.deg2rad(my_df['peak_wind_drct'])) * my_df['peak_wind_gust']
            my_df['u_peak'].attrs['units'] = 'm s-1'
            my_df['u_peak'].attrs['long_name'] = 'Zonal component of surface wind'
            my_df['v_peak'] = -np.cos(np.deg2rad(my_df['peak_wind_drct'])) * my_df['peak_wind_gust']
            my_df['v_peak'].attrs['units'] = 'm s-1'
            my_df['v_peak'].attrs['long_name'] = 'Meridional component of surface wind'
            my_df['metar'].attrs['long_name'] = 'Raw METAR code'
            my_df.attrs['_datastream'] = stations
            buf.close()

            station_obs[stations] = my_df
    return station_obs


def _download_data(uri):
    attempt = 0
    while attempt < 6:
        try:
            data = urlopen(uri, timeout=300).read().decode('utf-8')
            if data is not None and not data.startswith('ERROR'):
                return data
        except Exception as exp:
            print(f'download_data({uri}) failed with {exp}')
            time.sleep(5)
        attempt += 1

    print('Exhausted attempts to download, returning empty data')
    return ''



"""
This module contains the common procedures used by all modules of the ARM
Community Toolkit.

"""

from .data_utils import (
    ChangeUnits,
    accumulate_precip,
    add_in_nan,
    assign_coordinates,
    convert_units,
    create_pyart_obj,
    get_missing_value,
    ts_weighted_average,
)
from .datetime_utils import (
    dates_between,
    datetime64_to_datetime,
    determine_time_delta,
    numpy_to_arm_date,
    reduce_time_ranges,
    date_parser,
)
from .geo_utils import add_solar_variable, destination_azimuth_distance
from .inst_utils import decode_present_weather
from .qc_utils import calculate_dqr_times
from .radiance_utils import planck_converter
from .ship_utils import calc_cog_sog



"""
The procedures in this module contain corrections for various datasets.

"""

from .ceil import *  # noqa: F403
from .doppler_lidar import *  # noqa: F403
from .mpl import *  # noqa: F403
from .raman_lidar import *  # noqa: F403
from .ship import *  # noqa: F403



"""
This module contains procedures for working with QC information
and for applying tests to data.

"""

from .arm import *  # noqa: F403
from .clean import *  # noqa: F403
from .qcfilter import *  # noqa: F403
from .qctests import *  # noqa: F403
from .radiometer_tests import *  # noqa: F403
from .sp2 import SP2ParticleCriteria, get_waveform_statistics



"""
This module contains procedures for reading and writing various ARM datasets.

"""

from .armfiles import WriteDataset, check_arm_standards, create_obj_from_arm_dod, read_netcdf
from .csvfiles import read_csv
from .mpl import proc_sigma_mplv5_read, read_sigma_mplv5
from .noaagml import (
    read_gml,
    read_gml_co2,
    read_gml_halo,
    read_gml_met,
    read_gml_ozone,
    read_gml_radiation,
)
from .noaapsl import read_psl_wind_profiler
from .pysp2 import read_hk_file, read_sp2, read_sp2_dat



"""
Functions for calculated cloud base height that are instrument agnostic.

"""

import numpy as np
import xarray as xr
from scipy import ndimage


def generic_sobel_cbh(
    obj,
    variable=None,
    height_dim=None,
    var_thresh=None,
    fill_na=None,
    return_thresh=False,
):
    """
    Function for calculating cloud base height from lidar/radar data
    using a basic sobel filter and thresholding. Note, this was not
    initially based on any published work, but a lit review indicates
    that there have been similar methods employed to detect boundary
    layer heights.

    Parameters
    ----------
    obj : ACT Object
        ACT object where data are stored.
    variable : string
        Variable on which to process.
    height_dim : string
        Height variable to use for CBH values.
    var_thresh : float
        Thresholding for variable if needed.
    fill_na : float
        What to fill nans with in DataArray if any.

    Returns
    -------
    new_obj : ACT Object
        ACT Object with cbh values included as variable.

    Examples
    --------
    In testing on the ARM KAZR and MPL data, the following methods
    tended to work best for thresholding/corrections/etc.

    .. code-block:: python

        kazr = act.retrievals.cbh.generic_sobel_cbh(
            kazr, variable="reflectivity_copol", height_dim="range", var_thresh=-10.0
        )

        mpl = act.corrections.mpl.correct_mpl(mpl)
        mpl.range_bins.values = mpl.height.values[0, :] * 1000.0
        mpl.range_bins.attrs["units"] = "m"
        mpl["signal_return_co_pol"].values[:, 0:10] = 0.0
        mpl = act.retrievals.cbh.generic_sobel_cbh(
            mpl,
            variable="signal_return_co_pol",
            height_dim="range_bins",
            var_thresh=10.0,
            fill_na=0.0,
        )

        ceil = act.retrievals.cbh.generic_sobel_cbh(
            ceil,
            variable="backscatter",
            height_dim="range",
            var_thresh=1000.0,
            fill_na=0,
        )

    """
    if variable is None:
        return
    if fill_na is None:
        fill_na = var_thresh

    # Pull data into Standalone DataArray
    data = obj[variable]

    # Apply thresholds if set
    if var_thresh is not None:
        data = data.where(data.values > var_thresh)

    # Fill with fill_na values
    data = data.fillna(fill_na)

    # If return_thresh is True, replace variable data with
    # thresholded data
    if return_thresh is True:
        obj[variable].values = data.values

    # Apply Sobel filter to data and smooth the results
    data = data.values
    edge = ndimage.sobel(data)
    edge = ndimage.uniform_filter(edge, size=3, mode='nearest')

    # Create Data Array
    edge_obj = xr.DataArray(edge, dims=obj[variable].dims)

    # Filter some of the resulting edge data to get defined edges
    edge_obj = edge_obj.where(edge_obj > 5.0)
    edge_obj = edge_obj.fillna(fill_na)

    # Do a diff along the height dimension to define edge
    diff = edge_obj.diff(dim=1)

    # Get height variable to use for cbh
    height = obj[height_dim].values

    # Run through times and find the height
    cbh = []
    for i in range(np.shape(diff)[0]):
        index = np.where(diff[i, :] > 5.0)[0]
        if len(np.shape(height)) > 1:
            ht = height[i, :]
        else:
            ht = height

        if len(index) > 0:
            cbh.append(ht[index[0] + 1])
        else:
            cbh.append(np.nan)

    # Create DataArray to add to Object
    da = xr.DataArray(cbh, dims=['time'], coords=[obj['time'].values])
    obj['cbh_sobel'] = da
    obj['cbh_sobel'].attrs['long_name'] = ' '.join(
        ['CBH calculated from', variable, 'using sobel filter']
    )
    obj['cbh_sobel'].attrs['units'] = obj[height_dim].attrs['units']

    return obj



"""
Stores the class for ContourDisplay.

"""

import numpy as np
from scipy.interpolate import Rbf

# Import Local Libs
from .plot import Display


class ContourDisplay(Display):
    """
    This subclass contains routines that are specific to plotting
    contour plots from data. It is inherited from Display and therefore
    contains all of Display's attributes and methods.

    """

    def __init__(self, obj, subplot_shape=(1,), ds_name=None, **kwargs):
        super().__init__(obj, subplot_shape, ds_name, **kwargs)

    def create_contour(
        self,
        fields=None,
        time=None,
        function='cubic',
        subplot_index=(0,),
        contour='contourf',
        grid_delta=(0.01, 0.01),
        grid_buffer=0.1,
        twod_dim_value=None,
        **kwargs,
    ):
        """
        Extracts, grids, and creates a contour plot. If subplots have not been
        added yet, an axis will be created assuming that there is only going
        to be one plot.

        Parameters
        ----------
        fields : dict
            Dictionary of fields to use for x, y, and z data.
        time : datetime
            Time in which to slice through objects.
        function : string
            Defaults to cubic function for interpolation.
            See scipy.interpolate.Rbf for additional options.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.
        contour : str
            Whether to use contour or contourf as plotting function.
            Default is contourf
        grid_delta : 1D tuple, list, or array
            x and y deltas for creating grid.
        grid_buffer : float
            Buffer to apply to grid.
        twod_dim_value : float
            If the field is 2D, which dimension value to pull.
            I.e. if dim is depths of [5, 10, 50, 100] specifying 50
            would index the data at 50
        **kwargs : keyword arguments
            The keyword arguments for :func:`plt.contour`

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle of the plot.

        """
        # Get x, y, and z data by looping through each dictionary
        # item and extracting data from appropriate time
        x = []
        y = []
        z = []
        for ds in self._obj:
            obj = self._obj[ds]
            if ds not in fields:
                continue
            field = fields[ds]
            if obj[field[2]].sel(time=time).values.size > 1:
                dim_values = obj[obj[field[2]].dims[1]].values
                if twod_dim_value is None:
                    dim_index = 0
                else:
                    dim_index = np.where(dim_values == twod_dim_value)
                if dim_index[0].size == 0:
                    continue
                if np.isnan(obj[field[2]].sel(time=time).values[dim_index]):
                    continue
                z.append(obj[field[2]].sel(time=time).values[dim_index].tolist())
            else:
                if np.isnan(obj[field[2]].sel(time=time).values):
                    continue
                z.append(obj[field[2]].sel(time=time).values.tolist())

            if obj[field[0]].values.size > 1:
                x.append(obj[field[0]].sel(time=time).values.tolist())
            else:
                x.append(obj[field[0]].values.tolist())

            if obj[field[1]].values.size > 1:
                y.append(obj[field[1]].sel(time=time).values.tolist())
            else:
                y.append(obj[field[1]].values.tolist())

        # Create a meshgrid for gridding onto
        xs = np.arange(np.min(x) - grid_buffer, np.max(x) + grid_buffer, grid_delta[0])
        ys = np.arange(np.min(y) - grid_buffer, np.max(y) + grid_buffer, grid_delta[1])
        xi, yi = np.meshgrid(xs, ys)

        # Use scipy radial basis function to interpolate data onto grid
        rbf = Rbf(x, y, z, function=function)
        zi = rbf(xi, yi)

        # Create contour plot
        if contour == 'contour':
            self.contour(xi, yi, zi, subplot_index=subplot_index, **kwargs)
        elif 'contourf':
            self.contourf(xi, yi, zi, subplot_index=subplot_index, **kwargs)
        else:
            raise ValueError(
                "Invalid contour plot type. Please choose either 'contourf' or 'contour'"
            )

        return self.axes[subplot_index]

    def contourf(self, x, y, z, subplot_index=(0,), **kwargs):
        """
        Base function for filled contours if user already has data gridded.
        If subplots have not been added yet, an axis will be created
        assuming that there is only going to be one plot.

        Parameters
        ----------
        x : array-like
            x coordinates or grid for z.
        y : array-like
            y coordinates or grid for z.
        z : array-like(x,y)
            Values over which to contour.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.
        **kwargs : keyword arguments
            The keyword arguments for :func:`plt.contourf`

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle of the plot.

        """
        self.axes[subplot_index].contourf(x, y, z, **kwargs)

        return self.axes[subplot_index]

    def contour(self, x, y, z, subplot_index=(0,), **kwargs):
        """
        Base function for contours if user already has data gridded.
        If subplots have not been added yet, an axis will be created
        assuming that there is only going to be one plot.

        Parameters
        ----------
        x : array-like
            x coordinates or grid for z.
        y : array-like
            y coordinates or grid for z.
        z : array-like(x, y)
            Values over which to contour.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.
        **kwargs : keyword arguments
            The keyword arguments for :func:`plt.contour`

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle of the plot.

        """
        self.axes[subplot_index].contour(x, y, z, **kwargs)

        return self.axes[subplot_index]

    def plot_vectors_from_spd_dir(
        self,
        fields,
        time=None,
        subplot_index=(0,),
        mesh=False,
        function='cubic',
        grid_delta=(0.01, 0.01),
        grid_buffer=0.1,
        **kwargs,
    ):
        """
        Extracts, grids, and creates a contour plot.
        If subplots have not been added yet, an axis will be created
        assuming that there is only going to be one plot.

        Parameters
        ----------
        fields : dict
            Dictionary of fields to use for x, y, and z data.
        time : datetime
            Time in which to slice through objects.
        mesh : boolean
            Set to True to interpolate u and v to grid and create wind barbs.
        function : string
            Defaults to cubic function for interpolation.
            See scipy.interpolate.Rbf for additional options.
        grid_delta : 1D tuple, list, or array
            x and y deltas for creating grid.
        grid_buffer : float
            Buffer to apply to grid.
        **kwargs : keyword arguments
            The keyword arguments for :func:`plt.barbs`

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle of the plot.

        """
        # Get x, y, and z data by looping through each dictionary
        # item and extracting data from appropriate time
        x = []
        y = []
        wspd = []
        wdir = []
        for ds in self._obj:
            obj = self._obj[ds]
            field = fields[ds]
            if obj[field[0]].values.size > 1:
                x.append(obj[field[0]].sel(time=time).values.tolist())
            else:
                x.append(obj[field[0]].values.tolist())

            if obj[field[1]].values.size > 1:
                y.append(obj[field[1]].sel(time=time).values.tolist())
            else:
                y.append(obj[field[1]].values.tolist())
            wspd.append(obj[field[2]].sel(time=time).values.tolist())
            wdir.append(obj[field[3]].sel(time=time).values.tolist())

        # Calculate u and v
        tempu = -np.sin(np.deg2rad(wdir)) * wspd
        tempv = -np.cos(np.deg2rad(wdir)) * wspd

        if mesh is True:
            # Create a meshgrid for gridding onto
            xs = np.arange(min(x) - grid_buffer, max(x) + grid_buffer, grid_delta[0])
            ys = np.arange(min(y) - grid_buffer, max(y) + grid_buffer, grid_delta[1])
            xi, yi = np.meshgrid(xs, ys)

            # Use scipy radial basis function to interpolate data onto grid
            rbf = Rbf(x, y, tempu, function=function)
            u = rbf(xi, yi)

            rbf = Rbf(x, y, tempv, function=function)
            v = rbf(xi, yi)
        else:
            xi = x
            yi = y
            u = tempu
            v = tempv

        self.barbs(xi, yi, u, v, **kwargs)

        return self.axes[subplot_index]

    def barbs(self, x, y, u, v, subplot_index=(0,), **kwargs):
        """
        Base function for wind barbs. If subplots have not been added yet,
        an axis will be created assuming that there is only going to be
        one plot.

        Parameters
        ----------
        x : array-like
            x coordinates or grid for z.
        y : array-like
            y coordinates or grid for z.
        u : array-like
            U component of vector.
        v : array-like
            V component of vector.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.
        **kwargs : keyword arguments
            The keyword arguments for :func:`plt.barbs`

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle of the plot.

        """
        self.axes[subplot_index].barbs(x, y, u, v, **kwargs)

        return self.axes[subplot_index]

    def plot_station(self, fields, time=None, subplot_index=(0,), text_color='white', **kwargs):
        """
        Extracts, grids, and creates a contour plot. If subplots have not
        been added yet, an axis will be created assuming that there is only
        going to be one plot.

        Parameters
        ----------
        fields : dict
            Dictionary of fields to use for x, y, and z data.
        time : datetime
            Time in which to slice through objects.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.
        text_color : string
            Color for text.
        **kwargs : keyword arguments
            The keyword arguments for :func:`plt.plot`

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle of the plot.

        """
        # Get x, y, and data by looping through each dictionary
        # item and extracting data from appropriate time
        for ds in self._obj:
            obj = self._obj[ds]
            field = fields[ds]
            for i, f in enumerate(field):
                if i == 0:
                    if obj[f].values.size > 1:
                        x = obj[f].sel(time=time).values.tolist()
                    else:
                        x = obj[f].values.tolist()
                elif i == 1:
                    if obj[f].values.size > 1:
                        y = obj[f].sel(time=time).values.tolist()
                    else:
                        y = obj[f].values.tolist()
                    self.axes[subplot_index].plot(x, y, '*', **kwargs)
                else:
                    data = obj[f].sel(time=time).values.tolist()
                    offset = 0.02
                    if i == 2:
                        x1 = x - 3.0 * offset
                        y1 = y + offset
                    if i == 3:
                        x1 = x + offset
                        y1 = y + offset
                    if i == 4:
                        x1 = x + offset
                        y1 = y - 2.0 * offset
                    if i == 5:
                        x1 = x - 3.0 * offset
                        y1 = y - 2.0 * offset
                    if data < 5:
                        string = str(round(data, 1))
                    else:
                        string = str(round(data))
                    self.axes[subplot_index].text(x1, y1, string, color=text_color)

        return self.axes[subplot_index]



' Unit tests for the ACT retrievals module. ' ''

import glob
import numpy as np
import pytest
import xarray as xr

import act

try:
    import pysp2
    PYSP2_AVAILABLE = True
except ImportError:
    PYSP2_AVAILABLE = False


def test_get_stability_indices():
    sonde_ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_SONDE1)

    try:
        sonde_ds = act.retrievals.calculate_stability_indicies(
            sonde_ds, temp_name='tdry', td_name='dp', p_name='pres', rh_name='rh'
        )
        metpy = True
    except ImportError:
        metpy = False

    if metpy is True:
        np.testing.assert_allclose(
            sonde_ds['parcel_temperature'].values[0:5],
            [269.85, 269.745276, 269.678006, 269.622444, 269.572331],
            rtol=1e-5,
        )
        assert sonde_ds['parcel_temperature'].attrs['units'] == 'kelvin'
        np.testing.assert_almost_equal(sonde_ds['surface_based_cape'], 0.62, decimal=2)
        assert sonde_ds['surface_based_cape'].attrs['units'] == 'J/kg'
        assert sonde_ds['surface_based_cape'].attrs['long_name'] == 'Surface-based CAPE'
        np.testing.assert_almost_equal(sonde_ds['surface_based_cin'], 0.000, decimal=3)
        assert sonde_ds['surface_based_cin'].attrs['units'] == 'J/kg'
        assert sonde_ds['surface_based_cin'].attrs['long_name'] == 'Surface-based CIN'
        np.testing.assert_almost_equal(sonde_ds['most_unstable_cape'], 0.000, decimal=3)
        assert sonde_ds['most_unstable_cape'].attrs['units'] == 'J/kg'
        assert sonde_ds['most_unstable_cape'].attrs['long_name'] == 'Most unstable CAPE'
        np.testing.assert_almost_equal(sonde_ds['most_unstable_cin'], 0.000, decimal=3)
        assert sonde_ds['most_unstable_cin'].attrs['units'] == 'J/kg'
        assert sonde_ds['most_unstable_cin'].attrs['long_name'] == 'Most unstable CIN'

        np.testing.assert_almost_equal(sonde_ds['lifted_index'], 28.4, decimal=1)
        assert sonde_ds['lifted_index'].attrs['units'] == 'kelvin'
        assert sonde_ds['lifted_index'].attrs['long_name'] == 'Lifted index'
        np.testing.assert_equal(sonde_ds['level_of_free_convection'], np.array(np.nan))
        assert sonde_ds['level_of_free_convection'].attrs['units'] == 'hectopascal'
        assert sonde_ds['level_of_free_convection'].attrs['long_name'] == 'Level of free convection'
        np.testing.assert_almost_equal(
            sonde_ds['lifted_condensation_level_temperature'], -8.07, decimal=2
        )
        assert sonde_ds['lifted_condensation_level_temperature'].attrs['units'] == 'degree_Celsius'
        assert (
            sonde_ds['lifted_condensation_level_temperature'].attrs['long_name']
            == 'Lifted condensation level temperature'
        )
        np.testing.assert_almost_equal(
            sonde_ds['lifted_condensation_level_pressure'], 927.1, decimal=1
        )
        assert sonde_ds['lifted_condensation_level_pressure'].attrs['units'] == 'hectopascal'
        assert (
            sonde_ds['lifted_condensation_level_pressure'].attrs['long_name']
            == 'Lifted condensation level pressure'
        )
    sonde_ds.close()


def test_generic_sobel_cbh():
    ceil = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_CEIL1)

    ceil = ceil.resample(time='1min').nearest()
    ceil = act.retrievals.cbh.generic_sobel_cbh(
        ceil, variable='backscatter', height_dim='range', var_thresh=1000.0, fill_na=0
    )
    cbh = ceil['cbh_sobel'].values
    assert cbh[500] == 615.0
    assert cbh[1000] == 555.0
    ceil.close()


def test_calculate_precipitable_water():
    sonde_ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_SONDE1)
    assert sonde_ds['tdry'].units == 'C', 'Temperature must be in Celsius'
    assert sonde_ds['rh'].units == '%', 'Relative Humidity must be a percentage'
    assert sonde_ds['pres'].units == 'hPa', 'Pressure must be in hPa'
    pwv_data = act.retrievals.calculate_precipitable_water(
        sonde_ds, temp_name='tdry', rh_name='rh', pres_name='pres'
    )
    np.testing.assert_almost_equal(pwv_data, 0.8028, decimal=3)
    sonde_ds.close()


def test_doppler_lidar_winds():
    # Process a single file
    dl_ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_DLPPI)
    result = act.retrievals.doppler_lidar.compute_winds_from_ppi(dl_ds, intensity_name='intensity')
    assert np.round(np.nansum(result['wind_speed'].values)).astype(int) == 1570
    assert np.round(np.nansum(result['wind_direction'].values)).astype(int) == 32635
    assert result['wind_speed'].attrs['units'] == 'm/s'
    assert result['wind_direction'].attrs['units'] == 'degree'
    assert result['height'].attrs['units'] == 'm'
    dl_ds.close()

    # Process multiple files
    dl_ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_DLPPI_MULTI)
    del dl_ds['range'].attrs['units']
    result = act.retrievals.doppler_lidar.compute_winds_from_ppi(dl_ds)
    assert np.round(np.nansum(result['wind_speed'].values)).astype(int) == 64419
    assert np.round(np.nansum(result['wind_direction'].values)).astype(int) == 733627
    dl_ds.close()


def test_aeri2irt():
    aeri_ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_AERI)
    aeri_ds = act.retrievals.aeri.aeri2irt(aeri_ds)
    assert np.round(np.nansum(aeri_ds['aeri_irt_equiv_temperature'].values)).astype(int) == 17372
    np.testing.assert_almost_equal(
        aeri_ds['aeri_irt_equiv_temperature'].values[7], 286.081, decimal=3
    )
    np.testing.assert_almost_equal(
        aeri_ds['aeri_irt_equiv_temperature'].values[-10], 285.366, decimal=3
    )
    aeri_ds.close()
    del aeri_ds


def test_sst():
    obj = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_IRTSST)
    obj = act.retrievals.irt.sst_from_irt(obj)
    np.testing.assert_almost_equal(obj['sea_surface_temperature'].values[0], 278.901, decimal=3)
    np.testing.assert_almost_equal(obj['sea_surface_temperature'].values[-1], 279.291, decimal=3)
    assert np.round(np.nansum(obj['sea_surface_temperature'].values)).astype(int) == 6699
    obj.close()


def test_calculate_sirs_variable():
    sirs_object = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_SIRS)
    met_object = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_MET1)

    obj = act.retrievals.radiation.calculate_dsh_from_dsdh_sdn(sirs_object)
    assert np.isclose(np.nansum(obj['derived_down_short_hemisp'].values), 60350, atol=1)

    obj = act.retrievals.radiation.calculate_irradiance_stats(
        obj,
        variable='derived_down_short_hemisp',
        variable2='down_short_hemisp',
        threshold=60,
    )
    assert np.isclose(np.nansum(obj['diff_derived_down_short_hemisp'].values), 527, atol=1)
    assert np.isclose(np.nansum(obj['ratio_derived_down_short_hemisp'].values), 392, atol=1)

    obj = act.retrievals.radiation.calculate_net_radiation(obj, smooth=30)
    assert np.ceil(np.nansum(obj['net_radiation'].values)) == 21915
    assert np.ceil(np.nansum(obj['net_radiation_smoothed'].values)) == 22316

    obj = act.retrievals.radiation.calculate_longwave_radiation(
        obj,
        temperature_var='temp_mean',
        vapor_pressure_var='vapor_pressure_mean',
        met_obj=met_object,
    )
    assert np.ceil(obj['monteith_clear'].values[25]) == 239
    assert np.ceil(obj['monteith_cloudy'].values[30]) == 318
    assert np.ceil(obj['prata_clear'].values[35]) == 234

    new_obj = xr.merge([sirs_object, met_object], compat='override')
    obj = act.retrievals.radiation.calculate_longwave_radiation(
        new_obj, temperature_var='temp_mean', vapor_pressure_var='vapor_pressure_mean'
    )
    assert np.ceil(obj['monteith_clear'].values[25]) == 239
    assert np.ceil(obj['monteith_cloudy'].values[30]) == 318
    assert np.ceil(obj['prata_clear'].values[35]) == 234

    sirs_object.close()
    met_object.close()


def test_calculate_pbl_liu_liang():
    files = glob.glob(act.tests.sample_files.EXAMPLE_TWP_SONDE_20060121)
    files2 = glob.glob(act.tests.sample_files.EXAMPLE_SONDE1)
    files += files2
    files.sort()

    pblht = []
    pbl_regime = []
    for i, r in enumerate(files):
        obj = act.io.armfiles.read_netcdf(r)
        obj['tdry'].attrs['units'] = 'degree_Celsius'
        obj = act.retrievals.sonde.calculate_pbl_liu_liang(obj, smooth_height=10)
        pblht.append(float(obj['pblht_liu_liang'].values))
        pbl_regime.append(obj['pblht_regime_liu_liang'].values)

    assert pbl_regime == ['NRL', 'NRL', 'NRL', 'NRL', 'NRL']
    np.testing.assert_array_almost_equal(pblht, [847.5, 858.2, 184.8, 197.7, 443.2], decimal=1)

    obj = act.io.armfiles.read_netcdf(files[1])
    obj['tdry'].attrs['units'] = 'degree_Celsius'
    obj = act.retrievals.sonde.calculate_pbl_liu_liang(obj, land_parameter=False)
    np.testing.assert_almost_equal(obj['pblht_liu_liang'].values, 733.6, decimal=1)

    obj = act.io.armfiles.read_netcdf(files[-2:])
    obj['tdry'].attrs['units'] = 'degree_Celsius'
    with np.testing.assert_raises(ValueError):
        obj = act.retrievals.sonde.calculate_pbl_liu_liang(obj)

    obj = act.io.armfiles.read_netcdf(files[0])
    obj['tdry'].attrs['units'] = 'degree_Celsius'
    temp = obj['tdry'].values
    temp[10:20] = 19.0
    temp[0:10] = -10
    obj['tdry'].values = temp
    obj = act.retrievals.sonde.calculate_pbl_liu_liang(obj, land_parameter=False)
    assert obj['pblht_regime_liu_liang'].values == 'SBL'

    with np.testing.assert_raises(ValueError):
        obj2 = obj.where(obj['alt'] < 1000.0, drop=True)
        obj2 = act.retrievals.sonde.calculate_pbl_liu_liang(obj2, smooth_height=15)

    with np.testing.assert_raises(ValueError):
        obj2 = obj.where(obj['pres'] < 200.0, drop=True)
        obj2 = act.retrievals.sonde.calculate_pbl_liu_liang(obj2, smooth_height=15)

    with np.testing.assert_raises(ValueError):
        temp[0:5] = -40
        obj['tdry'].values = temp
        obj = act.retrievals.sonde.calculate_pbl_liu_liang(obj)

    obj = act.io.armfiles.read_netcdf(files[0])
    obj['tdry'].attrs['units'] = 'degree_Celsius'
    temp = obj['tdry'].values
    temp[20:50] = 100.0
    obj['tdry'].values = temp
    with np.testing.assert_raises(ValueError):
        obj = act.retrievals.sonde.calculate_pbl_liu_liang(obj)


@pytest.mark.skipif(not PYSP2_AVAILABLE, reason="PySP2 is not installed.")
def test_sp2_waveform_stats():
    my_sp2b = act.io.read_sp2(act.tests.EXAMPLE_SP2B)
    my_ini = act.tests.EXAMPLE_INI
    my_binary = act.qc.get_waveform_statistics(my_sp2b, my_ini, parallel=False)
    assert my_binary.PkHt_ch1.max() == 62669.4
    np.testing.assert_almost_equal(np.nanmax(my_binary.PkHt_ch0.values), 98708.92915295, decimal=1)
    np.testing.assert_almost_equal(np.nanmax(my_binary.PkHt_ch4.values), 54734.05714286, decimal=1)


@pytest.mark.skipif(not PYSP2_AVAILABLE, reason="PySP2 is not installed.")
def test_sp2_psds():
    my_sp2b = act.io.read_sp2(act.tests.EXAMPLE_SP2B)
    my_ini = act.tests.EXAMPLE_INI
    my_binary = act.qc.get_waveform_statistics(my_sp2b, my_ini, parallel=False)
    my_hk = act.io.read_hk_file(act.tests.EXAMPLE_HK)
    my_binary = act.retrievals.calc_sp2_diams_masses(my_binary)
    ScatRejectKey = my_binary['ScatRejectKey'].values
    assert np.nanmax(
        my_binary['ScatDiaBC50'].values[ScatRejectKey == 0]) < 1000.0
    my_psds = act.retrievals.process_sp2_psds(my_binary, my_hk, my_ini)
    np.testing.assert_almost_equal(my_psds['NumConcIncan'].max(), 0.95805343)



try:
    import pysp2

    PYSP2_AVAILABLE = True
except ImportError:
    PYSP2_AVAILABLE = False


def calc_sp2_diams_masses(input_ds, debug=True, factor=1.0, Globals=None):
    """
    Calculates the scattering and incandescence diameters/BC masses for each particle.

    Parameters
    ----------
    input_ds: ACT Dataset
        The ACT Dataset containing the processed SP2 data.
    debug: boolean
        If true, print out particle rejection statistics
    factor: float
        Multiply soot masses by this factor for AquaDag calibation. Use
        1.3 for NSA.
    Globals: act.qc.SP2ParticleCriteria structure or None
        DMTGlobals structure containing calibration coefficients. Set to
        None to use default values for MOSAiC.

    Returns
    -------
    output_ds: ACT Dataset
        The ACT Dataset containing the scattering/incadescence diameters.
    """
    if PYSP2_AVAILABLE:
        return pysp2.util.calc_diams_masses(input_ds, debug, factor, Globals)
    else:
        raise ModuleNotFoundError('PySP2 needs to be installed to use this feature.')


def process_sp2_psds(
    particle_ds, hk_ds, config_file, deltaSize=0.005, num_bins=199, avg_interval=10
):
    """
    Processes the Scattering and BC mass size distributions:

    Parameters
    ----------
    particle_ds: xarrray Dataset
        The xarray Dataset containing the particle statistics generated by
        act.retrievals.calc_sp2_diams_masses.
    hk_ds: xarray Dataset
        The xarray Dataset containing the housekeeping variables
    config_file: file_name
        Path to the .INI file.
    deltaSize: float
        The size distribution bin width in microns.
    num_bins: int
        The number of size bins
    avg_interval: int
        The time in seconds to average the concentrations into.

    Returns
    -------
    psd_ds: xarray Dataset
        The xarray Dataset containing the time-averaged particle statistics.
    """
    if PYSP2_AVAILABLE:
        config = pysp2.io.read_config(config_file)
        return pysp2.util.process_psds(
            particle_ds, hk_ds, config, deltaSize, num_bins, avg_interval
        )
    else:
        raise ModuleNotFoundError('PySP2 needs to be installed to use this feature.')



"""
Functions for doppler lidar specific retrievals

"""
import warnings

import numpy as np
import xarray as xr


def compute_winds_from_ppi(
    obj,
    elevation_name='elevation',
    azimuth_name='azimuth',
    radial_velocity_name='radial_velocity',
    snr_name='signal_to_noise_ratio',
    intensity_name=None,
    snr_threshold=0.008,
    remove_all_missing=False,
    condition_limit=1.0e4,
    return_obj=None,
):
    """
    This function will convert a Doppler Lidar PPI scan into vertical
    distribution of horizontal wind direction and speed.

    Code was adapted by Kenneth E Kehoe from code developed by Rob K
    Newsom. Please see the reference noted below and cite accordingly.

    Parameters
    ----------
    obj : Xarray Dataset Object
        The Dataset object containing PPI scan to be converte into winds.
    elevation_name : str
        The name of the elevation variable in the Dataset object.
    azimuth_name : str
        The name of the azimuth variable in the Dataset object.
    radial_velocity_name : str
        The name of the radial velocity variable in the Dataset object.
    snr_name : str
        The name of the signal to noise variable in the Dataset object.
    intensity_name : str
        The name of the intensity variable in the Dataset object. If this
        is set will use intensity instead of signal to noise ratio.
        variable.
    snr_threshold : float
        The signal to noise lower threshold used to decide which values to use.
    remove_all_missing : boolean
        Option to not add a time step in the returned object where all values
        are set to NaN
    condition_limit : float
        Upper limit used with Normalized data to check if data should be
        converted from scan signal to noise ration to wind speeds and
        directions.
    return_obj : None or  Xarray Dataset Object
        If set to a Xarray Dataset Object the calculated winds object will
        be concatinated onto this object. This is to allow looping over
        this function for many scans and returning a single object.

    Returns
    -------
    obj : Xarray Dataset Object or None
        The winds converted from PPI scan to horizontal wind speeds and wind
        directions along with wind speed error and wind direction error. If
        there is a problem determining the breaks between PPI scans, will
        return None.

    References
    ----------
    Rob K Newsom, Alan W Brewer, James M Wilczak, Daniel E Wolfe,
    Steven P Oncley and Julie K Lundquist ; Validating Precision Estimates in
    Horizontal Wind Measurements from a Doppler Lidar, Atmospheric Measurement
    Techniques Discussions 2016, 10, 1-30

    """
    azimuth = obj[azimuth_name].values
    azimuth_rounded = np.round(azimuth).astype(int)

    # Determine where the azimuth scans repeate to get range for each PPI
    index = np.where(azimuth_rounded == azimuth_rounded[0])[0]
    if index.size == 0:
        print(
            '\nERROR: Having trouble determining the PPI scan breaks '
            'in compute_winds_from_ppi().\n'
        )
        return return_obj

    if index.size == 1:
        num_scans = azimuth.size
    else:
        num_scans = index[1] - index[0]

    # Loop over each PPI scan
    for start_index in index:
        scan_index = range(start_index, start_index + num_scans)
        # Since this can run while instrument is making measurements
        # the number of PPI scans may not match exactly. This will
        # adjust the number of scans in case there is an issue.
        if scan_index[-1] > obj[elevation_name].values.size:
            scan_index = range(start_index, obj[elevation_name].values.size)

        elevation = np.radians(obj[elevation_name].values[scan_index])
        azimuth = np.radians(obj[azimuth_name].values[scan_index])
        doppler = obj[radial_velocity_name].values[scan_index]
        if intensity_name is not None:
            intensity = obj[intensity_name].values[scan_index, :]
            snr = intensity - 1
            del intensity
            snr_name = intensity_name
        else:
            try:
                snr = obj[snr_name].values[scan_index, :]
            except KeyError:
                intensity = obj['intensity'].values[scan_index, :]
                snr = intensity - 1
                del intensity
                snr_name = 'intensity'

        height_name = list(set(obj[snr_name].dims) - {'time'})[0]
        rng = obj[height_name].values
        try:
            height_units = obj[height_name].attrs['units']
        except KeyError:
            if rng[0] > 0:
                height_units = 'm'
            else:
                height_units = 'km'
        time = obj['time'].values[scan_index]

        height = rng * np.median(np.sin(elevation))
        xhat = np.sin(azimuth) * np.cos(elevation)
        yhat = np.cos(azimuth) * np.cos(elevation)
        zhat = np.sin(elevation)

        dims = snr.shape

        # mean_snr = np.nanmean(snr, axis=1)
        u_wind = np.full(dims[1], np.nan)
        v_wind = np.full(dims[1], np.nan)
        w_wind = np.full(dims[1], np.nan)
        u_err = np.full(dims[1], np.nan)
        v_err = np.full(dims[1], np.nan)
        w_err = np.full(dims[1], np.nan)
        residual = np.full(dims[1], np.nan)
        chisq = np.full(dims[1], np.nan)
        corr = np.full(dims[1], np.nan)

        # Loop over each level
        for ii in range(dims[1]):
            ur1 = doppler[:, ii]
            snr1 = snr[:, ii]
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', category=RuntimeWarning)
                index = np.where((snr1 >= snr_threshold) & np.isfinite(ur1))[0]
            count = index.size
            if count >= 4:
                ur1 = ur1[index]
                xhat1 = xhat[index]
                yhat1 = yhat[index]
                zhat1 = zhat[index]

                a = np.full((3, 3), np.nan)
                b = np.full(3, np.nan)

                a[0, 0] = np.sum(xhat1**2)
                a[1, 0] = np.sum(xhat1 * yhat1)
                a[2, 0] = np.sum(xhat1 * zhat1)

                a[0, 1] = a[1, 0]
                a[1, 1] = np.sum(yhat1**2)
                a[2, 1] = np.sum(yhat1 * zhat1)

                a[0, 2] = a[2, 0]
                a[1, 2] = a[2, 1]
                a[2, 2] = np.sum(zhat1**2)

                b[0] = np.sum(ur1 * xhat1)
                b[1] = np.sum(ur1 * yhat1)
                b[2] = np.sum(ur1 * zhat1)

                ainv = np.linalg.inv(a)
                condition = np.linalg.norm(a) * np.linalg.norm(ainv)  # Condition Number ?
                if condition < condition_limit:
                    c = b @ ainv
                    u_wind[ii] = c[0]
                    v_wind[ii] = c[1]
                    w_wind[ii] = c[2]
                    ur_fit = xhat1 * u_wind[ii] + yhat1 * v_wind[ii] + zhat1 * w_wind[ii]
                    chisq[ii] = np.sum((ur_fit - ur1) ** 2)
                    residual[ii] = np.sqrt(chisq[ii] / count)
                    with warnings.catch_warnings():
                        warnings.filterwarnings('ignore', category=RuntimeWarning)
                        corr[ii] = np.corrcoef(ur_fit, ur1)[0, 1]
                    u_err[ii] = np.sqrt((chisq[ii] / (count - 3)) * ainv[0, 0])
                    v_err[ii] = np.sqrt((chisq[ii] / (count - 3)) * ainv[1, 1])
                    w_err[ii] = np.sqrt((chisq[ii] / (count - 3)) * ainv[2, 2])

        # Compute windspeed and direction
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=RuntimeWarning)
            wspd = np.sqrt(u_wind**2 + v_wind**2)
            wdir = np.degrees(np.arctan2(u_wind, v_wind) + np.pi)

            wspd_err = np.sqrt((u_wind * u_err) ** 2 + (v_wind * v_err) ** 2) / wspd
            wdir_err = np.degrees(
                np.sqrt((u_wind * v_err) ** 2 + (v_wind * u_err) ** 2) / wspd**2
            )

        if remove_all_missing and np.isnan(wspd).all():
            continue

        time = time[0] + (time[-1] - time[0]) / 2
        time = time.reshape(
            1,
        )
        wspd = wspd.reshape(1, rng.size)
        wdir = wdir.reshape(1, rng.size)
        wspd_err = wspd_err.reshape(1, rng.size)
        wdir_err = wdir_err.reshape(1, rng.size)
        corr = corr.reshape(1, rng.size)
        residual = residual.reshape(1, rng.size)
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=RuntimeWarning)
            snr_mean = np.nanmean(snr, axis=0)
        snr_mean = snr_mean.reshape(1, rng.size)
        new_object = xr.Dataset(
            {
                'wind_speed': (
                    ('time', 'height'),
                    wspd,
                    {'long_name': 'Wind speed', 'units': 'm/s'},
                ),
                'wind_direction': (
                    ('time', 'height'),
                    wdir,
                    {'long_name': 'Wind direction', 'units': 'degree'},
                ),
                'wind_speed_error': (
                    ('time', 'height'),
                    wspd_err,
                    {'long_name': 'Wind direction error', 'units': 'm/s'},
                ),
                'wind_direction_error': (
                    ('time', 'height'),
                    wdir_err,
                    {'long_name': 'Wind direction error', 'units': 'degree'},
                ),
                'signal_to_noise_ratio': (
                    ('time', 'height'),
                    snr_mean,
                    {
                        'long_name': 'Signal to noise ratio mean over PPI scan',
                        'units': '1',
                    },
                ),
                'residual': (
                    ('time', 'height'),
                    residual,
                    {
                        'long_name': 'Residual values (Square Root of Chi Square)',
                        'units': 'm/s',
                    },
                ),
                'correlation': (
                    ('time', 'height'),
                    corr,
                    {'long_name': 'Correlation coefficient', 'units': '1'},
                ),
            },
            {
                'time': ('time', time, {'long_name': 'Time in UTC'}),
                'height': (
                    'height',
                    height,
                    {'long_name': 'Height to center of bin', 'units': height_units},
                ),
            },
        )

        if isinstance(return_obj, xr.core.dataset.Dataset):
            return_obj = xr.concat([return_obj, new_object], dim='time')
        else:
            return_obj = new_object

    return return_obj



"""
This module contains procedures for exploring and downloading data on
ARM Data Discovery.

"""

from .get_armfiles import download_data
from .get_asos import get_asos
from .get_cropscape import croptype



"""
Script for downloading data from ARM's Live Data Webservice

"""

import argparse
import json
import os
import sys

try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen

from act.utils import date_parser


def download_data(username, token, datastream, startdate, enddate, time=None, output=None):
    """
    This tool will help users utilize the ARM Live Data Webservice to download
    ARM data.

    Parameters
    ----------
    username : str
        The username to use for logging into the ADC archive.
    token : str
        The access token for accessing the ADC archive.
    datastream : str
        The name of the datastream to acquire.
    startdate : str
        The start date of the data to acquire. Formats accepted are
        YYYY-MM-DD, DD.MM.YYYY, DD/MM/YYYY, YYYYMMDD or YYYY/MM/DD.
    enddate : str
        The end date of the data to acquire. Formats accepted are
        YYYY-MM-DD, DD.MM.YYYY, DD/MM/YYYY, YYYYMMDD or YYYY/MM/DD.
    time: str or None
        The specific time. Format is HHMMSS. Set to None to download all files
        in the given date interval.
    output : str
        The output directory for the data. Set to None to make a folder in the
        current working directory with the same name as *datastream* to place
        the files in.

    Returns
    -------
    files : list
        Returns list of files retrieved

    Notes
    -----
    This programmatic interface allows users to query and automate
    machine-to-machine downloads of ARM data. This tool uses a REST URL and
    specific parameters (saveData, query), user ID and access token, a
    datastream name, a start date, and an end date, and data files matching
    the criteria will be returned to the user and downloaded.

    By using this web service, users can setup cron jobs and automatically
    download data from /data/archive into their workspace. This will also
    eliminate the manual step of following a link in an email to download data.
    All other data files, which are not on the spinning
    disk (on HPSS), will have to go through the regular ordering process.
    More information about this REST API and tools can be found on `ARM Live
    <https://adc.arm.gov/armlive/#scripts>`_.

    To login/register for an access token click `here
    <https://adc.arm.gov/armlive/livedata/home>`_.

    Author: Michael Giansiracusa
    Email: giansiracumt@ornl.gov

    Web Tools Contact: Ranjeet Devarakonda zzr@ornl.gov

    Examples
    --------
    This code will download the netCDF files from the sgpmetE13.b1 datastream
    and place them in a directory named sgpmetE13.b1. The data from 14 Jan to
    20 Jan 2017 will be downloaded. Replace *userName* and *XXXXXXXXXXXXXXXX*
    with your username and token for ARM Data Discovery. See the Notes for
    information on how to obtain a username and token.

    .. code-block:: python

        act.discovery.download_data(
            "userName", "XXXXXXXXXXXXXXXX", "sgpmetE13.b1", "2017-01-14", "2017-01-20"
        )

    """
    # default start and end are empty
    start, end = '', ''
    # start and end strings for query_url are constructed
    # if the arguments were provided
    if startdate:
        start = date_parser(startdate, output_format='%Y-%m-%d')
        start = f'&start={startdate}'
    if enddate:
        end = date_parser(enddate, output_format='%Y-%m-%d')
        end = f'&end={enddate}'
    # build the url to query the web service using the arguments provided
    query_url = (
        'https://adc.arm.gov/armlive/livedata/query?' + 'user={0}&ds={1}{2}{3}&wt=json'
    ).format(':'.join([username, token]), datastream, start, end)

    # get url response, read the body of the message,
    # and decode from bytes type to utf-8 string
    response_body = urlopen(query_url).read().decode('utf-8')
    # if the response is an html doc, then there was an error with the user
    if response_body[1:14] == '!DOCTYPE html':
        raise ConnectionRefusedError('Error with user. Check username or token.')

    # parse into json object
    response_body_json = json.loads(response_body)

    # construct output directory
    if output:
        # output files to directory specified
        output_dir = os.path.join(output)
    else:
        # if no folder given, add datastream folder
        # to current working dir to prevent file mix-up
        output_dir = os.path.join(os.getcwd(), datastream)

    # not testing, response is successful and files were returned
    if response_body_json is None:
        print('ARM Data Live Webservice does not appear to be functioning')
        return []

    num_files = len(response_body_json['files'])
    file_names = []
    if response_body_json['status'] == 'success' and num_files > 0:
        for fname in response_body_json['files']:
            if time is not None:
                if time not in fname:
                    continue
            print(f'[DOWNLOADING] {fname}')
            # construct link to web service saveData function
            save_data_url = (
                'https://adc.arm.gov/armlive/livedata/' + 'saveData?user={0}&file={1}'
            ).format(':'.join([username, token]), fname)
            output_file = os.path.join(output_dir, fname)
            # make directory if it doesn't exist
            if not os.path.isdir(output_dir):
                os.makedirs(output_dir)
            # create file and write bytes to file
            with open(output_file, 'wb') as open_bytes_file:
                open_bytes_file.write(urlopen(save_data_url).read())
            file_names.append(output_file)
    else:
        print(
            'No files returned or url status error.\n' 'Check datastream name, start, and end date.'
        )

    return file_names



"""
This module contains I/O operations for loading files that were created for the
Atmospheric Radiation Measurement program supported by the Department of Energy
Office of Science.
"""

import copy
import glob
import json
import re
import urllib
import warnings
from pathlib import Path

import numpy as np
import xarray as xr

import act.utils as utils


def read_netcdf(
    filenames,
    concat_dim=None,
    return_None=False,
    combine='by_coords',
    use_cftime=True,
    cftime_to_datetime64=True,
    combine_attrs='override',
    **kwargs,
):
    """

    Returns `xarray.Dataset` with stored data and metadata from a user-defined
    query of ARM-standard netCDF files from a single datastream. Has some procedures
    to ensure time is correctly fomatted in returned Dataset.

    Parameters
    ----------
    filenames : str or list
        Name of file(s) to read.
    concat_dim : str
        Dimension to concatenate files along. Default value is 'None'.
    return_None : bool, optional
        Catch IOError exception when file not found and return None.
        Default is False.
    combine : str
        String used by xarray.open_mfdataset() to determine how to combine
        data files into one Dataset. See Xarray documentation for options.
    use_cftime : boolean
        Option to use cftime library to parse the time units string and correctly
        establish the time values with a units string containing timezone offset.
        This will return the time in cftime format. See cftime_to_datetime64 if
        don't want to convert the times in xarray dataset from cftime to numpy datetime64.
    cftime_to_datetime64 : boolean
        If time is stored as cftime in xarray dataset convert to numpy datetime64. If time
        precision requried is sub millisecond set decode_times=False but leave
        cftime_to_datetime64=True. This will force it to use base_time and time_offset
        to set time.
    combine_attrs : str
        String indicating how to combine attrs of the objects being merged
    **kwargs : keywords
        Keywords to pass through to xarray.open_mfdataset().

    Returns
    -------
    act_obj : Object (or None)
        ACT dataset (or None if no data file(s) found).

    Examples
    --------
    This example will load the example sounding data used for unit testing.

    .. code-block :: python

        import act
        the_ds, the_flag = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_SONDE_WILDCARD)
        print(the_ds.attrs._datastream)

    """
    ds = None

    file_dates = []
    file_times = []

    # Add funciton keywords to kwargs dictionary for passing into open_mfdataset.
    kwargs['combine'] = combine
    kwargs['concat_dim'] = concat_dim
    kwargs['use_cftime'] = use_cftime
    kwargs['combine_attrs'] = combine_attrs

    # Create an exception tuple to use with try statements. Doing it this way
    # so we can add the FileNotFoundError if requested. Can add more error
    # handling in the future.
    except_tuple = (ValueError,)
    if return_None:
        except_tuple = except_tuple + (FileNotFoundError, OSError)

    try:
        # Read data file with Xarray function
        ds = xr.open_mfdataset(filenames, **kwargs)

    except except_tuple as exception:
        # If requested return None for File not found error
        if type(exception).__name__ == 'FileNotFoundError':
            return None

        # If requested return None for File not found error
        if type(exception).__name__ == 'OSError' and exception.args[0] == 'no files to open':
            return None

        # Look at error message and see if could be nested error message. If so
        # update combine keyword and try again. This should allow reading files
        # without a time variable but base_time and time_offset variables.
        if (
            kwargs['combine'] != 'nested'
            and type(exception).__name__ == 'ValueError'
            and exception.args[0] == 'Could not find any dimension coordinates '
            'to use to order the datasets for concatenation'
        ):
            kwargs['combine'] = 'nested'
            ds = xr.open_mfdataset(filenames, **kwargs)

        else:
            # When all else fails raise the orginal exception
            raise exception

    # Xarray has issues reading a CF formatted time units string if it contains
    # timezone offset without a [+|-] preceeding timezone offset.
    # https://github.com/pydata/xarray/issues/3644
    # To ensure the times are read in correctly need to set use_cftime=True.
    # This will read in time as cftime object. But Xarray uses numpy datetime64
    # natively. This will convert the cftime time values to numpy datetime64. cftime
    # does not preserve the time past ms precision. We will use ms precision for
    # the conversion.
    desired_time_precision = 'datetime64[ms]'
    for var_name in ['time', 'time_offset']:
        try:
            if (
                cftime_to_datetime64
                and 'time' in ds.dims
                and type(ds[var_name].values[0]).__module__.startswith('cftime.')
            ):
                # If we just convert time to datetime64 the group, sel, and other Xarray
                # methods will not work correctly because time is not indexed. Need to
                # use the formation of a Dataset to correctly set the time indexing.
                temp_ds = xr.Dataset(
                    {
                        var_name: (
                            ds[var_name].dims,
                            ds[var_name].values.astype(desired_time_precision),
                            ds[var_name].attrs,
                        )
                    }
                )
                ds[var_name] = temp_ds[var_name]
                temp_ds.close()

                # If time_offset is in file try to convert base_time as well
                if var_name == 'time_offset':
                    ds['base_time'].values = ds['base_time'].values.astype(desired_time_precision)
        except KeyError:
            pass

    # Check if "time" variable is not in the netCDF file. If so try to use
    # base_time and time_offset to make time variable. Basically a fix for incorrectly
    # formatted files. May require using decode_times=False to initially read the data.
    if (
        cftime_to_datetime64
        and 'time' in ds.dims
        and 'time' not in ds.coords
        and 'time_offset' in ds.data_vars
    ):
        try:
            ds = ds.rename({'time_offset': 'time'})
            ds = ds.set_coords('time')
            del ds['time'].attrs['units']
        except (KeyError, ValueError):
            pass

    # If "time" is not a datetime64 use base_time to calcualte corect values to datetime64
    # by adding base_time to time_offset. time_offset was renamed to time above.
    if (
        cftime_to_datetime64
        and 'time' in ds.dims
        and 'base_time' in ds.data_vars
        and not np.issubdtype(ds['time'].values.dtype, np.datetime64)
        and not type(ds['time'].values[0]).__module__.startswith('cftime.')
    ):
        # Use microsecond precision to create time since epoch. Then convert to datetime64
        if ds['base_time'].values == ds['time_offset'].values[0]:
            time = ds['time_offset'].values
        else:
            time = (ds['base_time'].values + ds['time_offset'].values * 1000000.0).astype(
                'datetime64[us]'
            )
        # Need to use a new Dataset creation to correctly index time for use with
        # .group and .resample methods in Xarray Datasets.
        temp_ds = xr.Dataset({'time': (ds['time'].dims, time, ds['time'].attrs)})

        ds['time'] = temp_ds['time']
        temp_ds.close()
        for att_name in ['units', 'ancillary_variables']:
            try:
                del ds['time'].attrs[att_name]
            except KeyError:
                pass

    # Adding support for wildcards
    if isinstance(filenames, str):
        filenames = glob.glob(filenames)

    # Get file dates and times that were read in to the object
    filenames.sort()
    for f in filenames:
        f = Path(f).name
        pts = re.match(r'(^[a-zA-Z0-9]+)\.([0-9a-z]{2})\.([\d]{8})\.([\d]{6})\.([a-z]{2,3}$)', f)
        # If Not ARM format, read in first time for info
        if pts is not None:
            pts = pts.groups()
            file_dates.append(pts[2])
            file_times.append(pts[3])
        else:
            if ds['time'].size > 1:
                dummy = ds['time'].values[0]
            else:
                dummy = ds['time'].values
            file_dates.append(utils.numpy_to_arm_date(dummy))
            file_times.append(utils.numpy_to_arm_date(dummy, returnTime=True))

    # Add attributes
    ds.attrs['_file_dates'] = file_dates
    ds.attrs['_file_times'] = file_times
    is_arm_file_flag = check_arm_standards(ds)

    # Ensure that we have _datastream set whether or no there's
    # a datastream attribute already.
    if is_arm_file_flag == 0:
        ds.attrs['_datastream'] = 'act_datastream'
    else:
        ds.attrs['_datastream'] = ds.attrs['datastream']

    ds.attrs['_arm_standards_flag'] = is_arm_file_flag

    return ds


def check_arm_standards(ds):
    """

    Checks to see if an xarray dataset conforms to ARM standards.
    Parameters
    ----------
    ds : xarray dataset
        The dataset to check.
    Returns
    -------
    flag : int
        The flag corresponding to whether or not the file conforms
        to ARM standards. Bit packed, so 0 for no, 1 for yes

    """
    the_flag = 1 << 0
    if 'datastream' not in ds.attrs.keys():
        the_flag = 0

    return the_flag


def create_obj_from_arm_dod(proc, set_dims, version='', fill_value=-9999.0, scalar_fill_dim=None):
    """

    Queries the ARM DOD api and builds an object based on the ARM DOD and
    the dimension sizes that are passed in.

    Parameters
    ----------
    proc : string
        Process to create the object off of. This is normally in the
        format of inst.level. i.e. vdis.b1 or kazrge.a1
    set_dims : dict
        Dictionary of dims from the DOD and the corresponding sizes.
        Time is required. Code will try and pull from DOD, unless set
        through this variable
        Note: names need to match exactly what is in the dod
        i.e. {'drop_diameter': 50, 'time': 1440}
    version : string
        Version number of the ingest to use. If not set, defaults to
        latest version
    fill_value : float
        Fill value for non-dimension variables. Dimensions cannot have
        duplicate values and are incrementally set (0, 1, 2)
    fill_value : str
        Depending on how the object is set up, sometimes the scalar values
        are dimensioned to the main dimension. i.e. a lat/lon is set to have
        a dimension of time. This is a way to set it up similarly.

    Returns
    -------
    obj : xarray Dataset
        ACT object populated with all variables and attributes.

    Examples
    --------
    .. code-block :: python

        dims = {'time': 1440, 'drop_diameter': 50}
        obj = act.io.armfiles.create_obj_from_arm_dod(
            'vdis.b1', dims, version='1.2', scalar_fill_dim='time')

    """
    # Set base url to get DOD information
    base_url = 'https://pcm.arm.gov/pcm/api/dods/'

    # Get data from DOD api
    with urllib.request.urlopen(base_url + proc) as url:
        data = json.loads(url.read().decode())

    # Check version numbers and alert if requested version in not available
    keys = list(data['versions'].keys())
    if version not in keys:
        warnings.warn(
            ' '.join(
                ['Version:', version, 'not available or not specified. Using Version:', keys[-1]]
            ),
            UserWarning,
        )
        version = keys[-1]

    # Create empty xarray dataset
    obj = xr.Dataset()

    # Get the global attributes and add to dataset
    atts = {}
    for a in data['versions'][version]['atts']:
        if a['name'] == 'string':
            continue
        if a['value'] is None:
            a['value'] = ''
        atts[a['name']] = a['value']

    obj.attrs = atts

    # Get variable information and create dataarrays that are
    # then added to the dataset
    # If not passed in through set_dims, will look to the DOD
    # if not set in the DOD, then will raise error
    variables = data['versions'][version]['vars']
    dod_dims = data['versions'][version]['dims']
    for d in dod_dims:
        if d['name'] not in list(set_dims.keys()):
            if d['length'] > 0:
                set_dims[d['name']] = d['length']
            else:
                raise ValueError(
                    'Dimension length not set in DOD for '
                    + d['name']
                    + ', nor passed in through set_dim'
                )
    for v in variables:
        dims = v['dims']
        dim_shape = []
        # Using provided dimension data, fill array accordingly for easy overwrite
        if len(dims) == 0:
            if scalar_fill_dim is None:
                data_na = fill_value
            else:
                data_na = np.full(set_dims[scalar_fill_dim], fill_value)
                v['dims'] = scalar_fill_dim
        else:
            for d in dims:
                dim_shape.append(set_dims[d])
            if len(dim_shape) == 1 and v['name'] == dims[0]:
                data_na = np.arange(dim_shape[0])
            else:
                data_na = np.full(dim_shape, fill_value)

        # Get attribute information. Had to do some things to get to print to netcdf
        atts = {}
        str_flag = False
        for a in v['atts']:
            if a['name'] == 'string':
                str_flag = True
                continue
            if a['value'] is None:
                continue
            if str_flag and a['name'] == 'units':
                continue
            atts[a['name']] = a['value']

        da = xr.DataArray(data=data_na, dims=v['dims'], name=v['name'], attrs=atts)
        obj[v['name']] = da

    return obj


@xr.register_dataset_accessor('write')
class WriteDataset:
    """

    Class for cleaning up Dataset before writing to file.

    """

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def write_netcdf(
        self,
        cleanup_global_atts=True,
        cleanup_qc_atts=True,
        join_char='__',
        make_copy=True,
        cf_compliant=False,
        delete_global_attrs=['qc_standards_version', 'qc_method', 'qc_comment'],
        FillValue=-9999,
        cf_convention='CF-1.8',
        **kwargs,
    ):
        """

        This is a wrapper around Dataset.to_netcdf to clean up the Dataset before
        writing to disk. Some things are added to global attributes during ACT reading
        process, and QC variables attributes are modified during QC cleanup process.
        This will modify before writing to disk to better
        match Climate & Forecast standards.

        Parameters
        ----------
        cleanup_global_atts : boolean
            Option to cleanup global attributes by removing any global attribute
            that starts with an underscore.
        cleanup_qc_atts : boolean
            Option to convert attributes that would be written as string array
            to be a single character string. CF 1.7 does not allow string attribures.
            Will use a single space a delimeter between values and join_char to replace
            white space between words.
        join_char : str
            The character sting to use for replacing white spaces between words when converting
            a list of strings to single character string attributes.
        make_copy : boolean
            Make a copy before modifying Dataset to write. For large Datasets this
            may add processing time and memory. If modifying the Dataset is OK
            try setting to False.
        cf_compliant : boolean
            Option to output file with additional attributes to make file Climate & Forecast
            complient. May require runing .clean.cleanup() method on the object to fix other
            issues first. This does the best it can but it may not be truely complient. You
            should read the CF documents and try to make complient before writing to file.
        delete_global_attrs : list
            Optional global attributes to be deleted. Defaults to some standard
            QC attributes that are not needed. Can add more or set to None to not
            remove the attributes.
        FillValue : int, float
            The value to use as a _FillValue in output file. This is used to fix
            issues with how Xarray handles missing_value upon reading. It's confusing
            so not a perfect fix. Set to None to leave Xarray to do what it wants.
            Set to a value to be the value used as _FillValue in the file and data
            array. This should then remove missing_value attribute from the file as well.
        cf_convention : str
            The Climate and Forecast convention string to add to Conventions attribute.
        **kwargs : keywords
            Keywords to pass through to Dataset.to_netcdf()

        Examples
        --------
        .. code-block :: python

            ds_object.write.write_netcdf(path='output.nc')

        """

        if make_copy:
            write_obj = copy.deepcopy(self._obj)
        else:
            write_obj = self._obj

        encoding = {}
        if cleanup_global_atts:
            for attr in list(write_obj.attrs):
                if attr.startswith('_'):
                    del write_obj.attrs[attr]

        if cleanup_qc_atts:
            check_atts = ['flag_meanings', 'flag_assessments']
            for var_name in list(write_obj.data_vars):
                if 'standard_name' not in write_obj[var_name].attrs.keys():
                    continue
                for attr_name in check_atts:
                    try:
                        if isinstance(write_obj[var_name].attrs[attr_name], (list, tuple)):
                            att_values = write_obj[var_name].attrs[attr_name]
                            for ii, att_value in enumerate(att_values):
                                att_values[ii] = att_value.replace(' ', join_char)

                            write_obj[var_name].attrs[attr_name] = ' '.join(att_values)
                    except KeyError:
                        pass

                # Tell .to_netcdf() to not add a _FillValue attribute for
                # quality control variables.
                if FillValue is not None:
                    encoding[var_name] = {'_FillValue': None}

            # Clean up _FillValue vs missing_value mess by creating an
            # encoding dictionary with each variable's _FillValue set to
            # requested fill value. May need to improve upon this for data type
            # and other issues in the future.
            if FillValue is not None:
                skip_variables = ['base_time', 'time_offset', 'qc_time'] + list(encoding.keys())
                for var_name in list(write_obj.data_vars):
                    if var_name not in skip_variables:
                        encoding[var_name] = {'_FillValue': FillValue}

        if delete_global_attrs is not None:
            for attr in delete_global_attrs:
                try:
                    del write_obj.attrs[attr]
                except KeyError:
                    pass

        # If requested update global attributes and variables attributes for required
        # CF attributes.
        if cf_compliant:
            # Get variable names and standard name for each variable
            var_names = list(write_obj.keys())
            standard_names = []
            for var_name in var_names:
                try:
                    standard_names.append(write_obj[var_name].attrs['standard_name'])
                except KeyError:
                    standard_names.append(None)

            # Check if time varible has axis and standard_name attribute
            coord_name = 'time'
            try:
                write_obj[coord_name].attrs['axis']
            except KeyError:
                try:
                    write_obj[coord_name].attrs['axis'] = 'T'
                except KeyError:
                    pass

            try:
                write_obj[coord_name].attrs['standard_name']
            except KeyError:
                try:
                    write_obj[coord_name].attrs['standard_name'] = 'time'
                except KeyError:
                    pass

            # Try to determine type of dataset by coordinate dimention named time
            # and other factors
            try:
                write_obj.attrs['FeatureType']
            except KeyError:
                dim_names = list(write_obj.dims)
                FeatureType = None
                if dim_names == ['time']:
                    FeatureType = 'timeSeries'
                elif len(dim_names) == 2 and 'time' in dim_names and 'bound' in dim_names:
                    FeatureType = 'timeSeries'
                elif len(dim_names) >= 2 and 'time' in dim_names:
                    for var_name in var_names:
                        dims = list(write_obj[var_name].dims)
                        if len(dims) == 2 and 'time' in dims:
                            prof_dim = list(set(dims) - {'time'})[0]
                            if write_obj[prof_dim].values.size > 2:
                                FeatureType = 'timeSeriesProfile'
                                break

                if FeatureType is not None:
                    write_obj.attrs['FeatureType'] = FeatureType

            # Add axis and positive attributes to variables with standard_name
            # equal to 'altitude'
            alt_variables = [
                var_names[ii] for ii, sn in enumerate(standard_names) if sn == 'altitude'
            ]
            for var_name in alt_variables:
                try:
                    write_obj[var_name].attrs['axis']
                except KeyError:
                    write_obj[var_name].attrs['axis'] = 'Z'

                try:
                    write_obj[var_name].attrs['positive']
                except KeyError:
                    write_obj[var_name].attrs['positive'] = 'up'

            # Check if the Conventions global attribute lists the CF convention
            try:
                Conventions = write_obj.attrs['Conventions']
                Conventions = Conventions.split()
                cf_listed = False
                for ii in Conventions:
                    if ii.startswith('CF-'):
                        cf_listed = True
                        break
                if not cf_listed:
                    Conventions.append(cf_convention)
                write_obj.attrs['Conventions'] = ' '.join(Conventions)

            except KeyError:
                write_obj.attrs['Conventions'] = str(cf_convention)

            # Reorder global attributes to ensure history is last
            try:
                global_attrs = write_obj.attrs
                history = copy.copy(global_attrs['history'])
                del global_attrs['history']
                global_attrs['history'] = history
            except KeyError:
                pass

        write_obj.to_netcdf(encoding=encoding, **kwargs)
